# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO refactor this module"""

from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import xarray as xr
from jax.tree_util import tree_flatten, tree_unflatten

from wax.gym.callbacks import Callback
from wax.modules.gym_feedback import GymOutput
from wax.stream import DatasetSchema


class Record(Callback):
    """Record rewards while running gym_unroll.

    Parameters
    ----------

    schema: Dataset schema : contains coords and encoders.
    format_dims : nested data structure with specification of dims for dataarray formatting.

    return_obs
        if true, also record env observations.

    return_action
        if true, also record agent actions.

    Methods
    -------

    get_outputs

    """

    __slots__ = [
        "schema",
        "format_dims",
        "return_obs",
        "return_action",
        "GymOutput",
        "format_outputs",
        "_sequence",
        "_treedef",
        "_times",
        "_dates",
        "_env",
    ]

    def __init__(
        self,
        schema: Optional[DatasetSchema] = None,
        format_dims: Any = None,
        return_obs: bool = False,
        return_action: bool = False,
        format_outputs: Any = None,
    ):
        self.schema = schema
        self.format_dims = format_dims
        self.return_obs = return_obs
        self.return_action = return_action
        self.format_outputs = format_outputs
        # state
        # same format as Gym module.
        self.GymOutput = GymOutput(self.return_obs, self.return_action)
        self._sequence = None
        self._treedef = None
        self._times: List = []
        self._dates: List = []
        self._env: Dict = {}

    def _on_step(self, env, agent, gym_state):
        if gym_state.done:
            return
        if gym_state.action is not None:

            outputs = self.GymOutput.format(
                gym_state.rw, gym_state.obs, gym_state.action
            )
            outputs, treedef = tree_flatten(outputs)
            if self._sequence is None:
                self._treedef = treedef
                self._sequence = tuple([] for _ in outputs)

            for seq_, output_ in zip(self._sequence, outputs):
                seq_.append(output_)

        self._env = env

    def get_outputs(self):
        """Get gym output sequence.

        Returns :
            unrolled output : reward recorded running gym_unroll.
                if return_obs is true, also return recorded env observations.
                if return_action is true, aslo return recorded agent actions.

        """
        sequence = self._sequence

        def _stack(x):
            if isinstance(x[0], (pd.DataFrame, pd.Series)):
                return pd.concat(x, axis=0)
            elif isinstance(x[0], (xr.Dataset, xr.DataArray)):
                return xr.concat(x, dim=x[0].dims[0])
            else:
                return np.stack(x, axis=0)

        sequence = list(map(_stack, sequence))
        sequence = tree_unflatten(self._treedef, sequence)
        return sequence



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial, reduce

import jax
import jax.numpy as jnp
import numpy as onp
import pandas as pd
import xarray as xr
from haiku import transform_with_state
from jax.numpy import array as DeviceArray
from jax.numpy import int32
from jax.tree_util import tree_flatten, tree_map
from numpy import array
from pandas import NaT

from wax.datasets.generate_temperature_data import (
    generate_temperature_data_multi_time_scale,
)
from wax.encode import encode_dataset
from wax.stream import (
    Stream,
    dataset_to_numpy,
    get_time_dataset,
    split_dataset_from_time_dims,
    tree_access_data,
    unroll_stream,
)
from wax.testing import assert_tree_all_close
from wax.unroll import unroll
from wax.utils import dict_map


def onp_half_precision(x):
    if x.dtype == onp.float64:
        return x.astype(onp.float32)
    if x.dtype == onp.int64:
        return x.astype(onp.int32)
    return x


def test_onp_half_precision():
    x = onp.array(1, dtype=onp.int32)
    assert onp_half_precision(x) == x


def generate_news_data():
    days = 2
    bins = 1000
    onp.random.seed(3)
    dates = pd.date_range(start="2000/01/02", periods=days, freq="B")
    news_times = pd.Index(
        reduce(
            lambda x, y: x + y,
            [
                (date + pd.timedelta_range(start=1, freq="1min", periods=bins)).tolist()
                for date in dates
            ],
        )
    )
    idx_news = onp.random.choice(len(news_times), 10)
    news_times = news_times[sorted(idx_news)]

    # create a bin with 5 observations
    news_times = news_times.tolist()
    news_times.append(news_times[0] + onp.timedelta64(3, "s"))
    news_times.append(news_times[0] + onp.timedelta64(10, "s"))
    news_times.append(news_times[0] + onp.timedelta64(20, "s"))
    news_times.append(news_times[0] + onp.timedelta64(27, "s"))
    news_times = pd.Index(sorted(news_times)).values
    news_times = news_times + onp.timedelta64(50, "ms")
    T = len(news_times)
    news_data = onp.random.normal(size=(T, 2)).astype(str)
    news_data = xr.DataArray(
        news_data,
        dims=("news_time", "news_attr"),
        coords={"news_time": news_times, "news_attr": ["head", "text"]},
    )
    return news_data


def prepare_test_data():
    xdata = generate_temperature_data_multi_time_scale()

    xdata["NEWS"] = generate_news_data()

    xdata["param"] = 1

    return xdata


def check_time_index(index, direct=True):
    if direct:
        name_news = "news_time_index"
        name_air = "day_index"
    else:
        name_news = "NEWS"
        name_air = "air"
    idx = index["news_time"].value[name_news]
    assert len(onp.where(idx > 0)[0].flatten()) == 27

    ref = index["news_time"].time.tolist()
    assert index["news_time"].time.shape == (124, 3)

    assert ref[0] == [NaT, NaT, NaT]
    assert ref[53] == [946872970050000001, 946872980050000001, 946872987050000001]

    ref = index["news_time"].value[name_news].tolist()
    assert ref[0] == [-99999, -99999, -99999]
    assert ref[53] == [2, 3, 4]

    assert index["news_time"].value[name_news].shape == (124, 3)

    ref = index["day"].time.tolist()[20:30]
    assert ref == [
        NaT,
        NaT,
        NaT,
        NaT,
        946684800000000000,
        946684800000000000,
        946684800000000000,
        946684800000000000,
        946684800000000000,
        946684800000000000,
    ]

    ref = index["day"].value[name_air].tolist()[20:30]
    assert (ref == array([-99999, -99999, -99999, -99999, 0, 0, 0, 0, 0, 0])).all()


def check_outputs(outputs, schema):

    # TODO: check what is correct...
    # assert outputs["NEWS"].shape == (25, 3, 2)
    assert outputs["NEWS"].shape == (124, 3)
    assert_tree_all_close(
        outputs["NEWS"][-1],
        DeviceArray([-99999, -99999, -99999], dtype=int32),
    )
    assert_tree_all_close(
        outputs["NEWS"][53],
        DeviceArray([27, 6, 23], dtype=int32),
    )

    assert_tree_all_close(
        outputs["NEWS"][:, 0][60:70],
        DeviceArray(
            [-99999, -99999, 23, -99999, 10, 1, -99999, -99999, -99999, -99999],
            dtype=int32,
        ),
    )

    # decode outputs
    assert outputs["NEWS"].shape == (124, 3)


def test_stream_dataset_dynamic_unroll2():
    dataset = prepare_test_data()
    assert dataset["NEWS"].shape == (14, 2)
    schema = Stream.get_dataset_schema(dataset, check=True)
    ffills = {"day": True}
    freqs = {"day": "d"}
    local_time = "time"
    buffer_maxlen = {"news_time": 3}

    datasets = split_dataset_from_time_dims(dataset)
    assert datasets.keys() == {"day", "news_time", "time"}
    assert datasets["news_time"]["NEWS"].shape == (14, 2)

    stream = Stream(
        local_time,
        freqs,
        ffills,
        buffer_maxlen,
        verbose=True,
        pbar=["time"],
        trace=True,
    )
    streams = {}
    for time_dim, ds in datasets.items():
        streams[time_dim] = stream.start_dataset_stream(ds, time_dim=time_dim)
    trace = stream.merge(streams)
    index = unroll_stream(trace)

    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, onp.ndarray)

    check_type(index)

    check_time_index(index, False)

    # convert to dict of numpy

    np_data = dict_map(dataset_to_numpy, datasets)

    # encode values
    assert datasets["news_time"]["NEWS"].shape == (14, 2)
    np_data = dict_map(partial(encode_dataset, schema.encoders), np_data)
    assert np_data["news_time"]["NEWS"].shape == (28,)
    # drop times from index structure
    np_index = {dim: var.value for dim, var in index.items()}

    # explicitly convert in np.float32 and int32 berfore jax conversion to avoid jax warnings.
    np_data, np_index = tree_map(onp_half_precision, (np_data, np_index))
    # convert to jax
    np_data, np_index = jax.device_put((np_data, np_index))

    # check data types:
    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, jnp.DeviceArray)

    check_type(np_data)
    check_type(np_index)

    access_dataset = transform_with_state(partial(tree_access_data, np_data, np_index))

    xs = onp.arange(len(schema.coords["time"]))

    # outputs, state = static_unroll(access_dataset, xs, next(seq))
    # now test dynamic unroll
    outputs, state = unroll(access_dataset, return_final_state=True)(xs)

    check_outputs(outputs["news_time"], schema)


def test_stream_dataset_dynamic_unroll3():
    dataset = prepare_test_data()
    assert dataset["NEWS"].shape == (14, 2)
    schema = Stream.get_dataset_schema(dataset, check=True)
    ffills = {"day": True}
    freqs = {"day": "d"}
    local_time = "time"
    buffer_maxlen = {"news_time": 3}

    time_dataset = get_time_dataset(dataset)

    time_datasets = split_dataset_from_time_dims(time_dataset)
    stream = Stream(
        local_time, freqs, ffills, buffer_maxlen, verbose=["time"], pbar=True
    )
    streams = stream.start_dataset_streams(time_datasets)
    trace = stream.merge(streams)
    time_index = unroll_stream(trace)

    # check time_index
    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, onp.ndarray)

    check_type(time_index)
    check_time_index(time_index)

    from .stream import get_dataset_index, get_dataset_index_from_stream_index

    # now convert in index for the original dataset.
    time_dataset_index = get_dataset_index_from_stream_index(time_index)
    dataset_index = get_dataset_index(dataset, time_dataset_index)
    assert dataset["param"].shape == ()

    assert dataset_index["param"].shape == (124,)
    # check that dataset_index and dataset have same variables.
    assert dataset.keys() == dataset_index.keys()

    # convert to dict of numpy

    np_data = dataset_to_numpy(dataset)
    np_index = dataset_to_numpy(dataset_index)
    assert np_index["param"].shape == (124,)

    # encode values
    assert dataset["NEWS"].shape == (14, 2)
    np_data_encoded = encode_dataset(schema.encoders, np_data)
    from wax.encode import decode_dataset

    np_data_reencoded = decode_dataset(schema.encoders, np_data_encoded)
    for key in np_data:
        assert (np_data[key] == np_data_reencoded[key]).all()
    np_data = np_data_encoded

    assert np_data["NEWS"].shape == (28,)
    # drop times from index structure

    # explicitly convert in np.float32 and int32 berfore jax conversion to avoid jax warnings.
    np_data, np_index = tree_map(onp_half_precision, (np_data, np_index))
    # convert to jax
    np_data, np_index = jax.device_put(
        (np_data, np_index),
    )

    # check data types:
    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, jnp.DeviceArray)

    check_type(np_data)
    check_type(np_index)
    assert np_index["param"].shape == (124,)
    access_dataset = transform_with_state(partial(tree_access_data, np_data, np_index))

    xs = onp.arange(len(schema.coords["time"]))

    # outputs, state = static_unroll(access_dataset, xs, next(seq))
    # now test dynamic unroll
    outputs, state = unroll(access_dataset, return_final_state=True)(xs)

    # decode outputs
    assert outputs["NEWS"].shape == (124, 3)
    check_outputs(outputs, schema)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime

import jax.numpy as jnp
import numpy as onp
import pandas as pd
import pytest

from wax.encode import (
    decode_datetime,
    decode_datetime64,
    decode_int64,
    decode_timestamp,
    encode_datetime,
    encode_datetime64,
    encode_int64,
    encode_timestamp,
    floor_datetime,
    floor_jax_datekey,
    less,
)


@pytest.mark.parametrize("seed", [10.0])
def test_encode_float(seed):
    with pytest.raises(TypeError):
        encode_int64(seed)


@pytest.mark.parametrize(
    "seed",
    [
        2**63 - 1,
        2**50 - 46,
        2**34,
        onp.int32(10),
        onp.uint32(10),
        onp.int64(10),
        10,
    ],
)
def test_encode(seed):
    i32 = encode_int64(seed)
    assert i32 is not seed
    assert i32.dtype == onp.uint32
    assert i32.shape == (2,)
    i64 = decode_int64(i32)
    assert i64 == seed


@pytest.mark.parametrize("seed", [onp.int32(10), onp.int64(10), 10])
def test_encode_int32(seed):
    i32 = encode_int64(seed)
    assert i32 is not seed
    assert i32.dtype == onp.uint32
    assert i32.shape == (2,)
    high, low = i32
    assert high == 0
    assert low == 10
    i64 = decode_int64(i32)
    assert i64 == seed


def test_encode_timestamp():
    date = pd.to_datetime(datetime.datetime.now())
    code = encode_timestamp(date)
    decoded_date = decode_timestamp(code)
    assert type(decoded_date) == type(date)
    assert decoded_date == date


def test_encode_datetime():
    date = datetime.datetime.now()
    code = encode_datetime(date)
    decoded_date = decode_datetime(code)
    assert type(decoded_date) == type(date)
    assert decoded_date == date


@pytest.mark.parametrize("dtype", ["<M8[us]", "<M8[ns]"])
def test_encode_datetime64_simple(dtype):
    date = onp.datetime64(datetime.datetime.now()).astype(dtype)
    code = encode_datetime64(date)
    decoded_date = decode_datetime64(code)
    assert type(decoded_date) == type(date)
    assert decoded_date == date


TEST_INT_LIST = [
    (1, 2),
    (10, 2**33),
    (2**32, 2**32 + 1),
    (2**32 - 1, 2**32),
]

# inverse order of t1, t2
TEST_INT_LIST += [(t2, t1) for t1, t2 in TEST_INT_LIST]


@pytest.mark.parametrize("t1, t2", TEST_INT_LIST)
def test_less(t1, t2):
    t1_is_less_t2 = t1 < t2
    ct1, ct2 = map(encode_int64, (t1, t2))
    assert less(ct1, ct2) == t1_is_less_t2

    # also check encode / decode
    assert decode_int64(ct1) == t1
    assert decode_int64(ct2) == t2


@pytest.mark.parametrize("dtype", ["<M8[us]", "<M8[ns]"])
def test_encode_datetime64(dtype):
    date = onp.datetime64(datetime.datetime.now()).astype(dtype)
    code = encode_datetime64(date)
    code = jnp.array(code)  # convert to jax DeviceArray
    assert isinstance(code, jnp.DeviceArray)

    # check decode
    decoded_date = decode_datetime64(code)
    # we are returned to numpy datetime
    assert type(decoded_date) == type(date)
    assert decoded_date == date

    # now check projection to date
    code_proj = floor_jax_datekey(code)
    date_proj = decode_datetime64(onp.array(code_proj))
    assert date_proj == floor_datetime(date)


def test_encode_datetime64_raises():
    with pytest.raises(TypeError):
        encode_datetime64(onp.array(2.0))

    with pytest.raises(TypeError):
        encode_datetime(onp.array(2.0))

    with pytest.raises(TypeError):
        encode_int64(onp.array(2.0))

    with pytest.raises(TypeError):
        encode_int64(onp.eye(2))

    with pytest.raises(TypeError):
        encode_timestamp(onp.eye(2))



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.3
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# +
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html
# -

# %matplotlib inline


import io
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, NamedTuple, Optional, TypeVar

import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import numpy as onp
import optax
import pandas as pd
import plotnine as gg
import requests
from sklearn.preprocessing import MinMaxScaler
from tqdm.auto import tqdm

from wax.accessors import register_wax_accessors
from wax.compile import jit_init_apply
from wax.encode import Encoder
from wax.modules import Buffer, FillNanInf, Lag, RollingMean
from wax.unroll import unroll

print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# # 🔭 Reconstructing the light curve of stars with LSTM 🔭
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/05_reconstructing_the_light_curve_of_stars.ipynb)

# Let's take a walk through the stars...
#
# This notebook is based on the study done in
# [this post by Christophe Pere](https://towardsdatascience.com/how-to-use-deep-learning-for-time-series-forecasting-3f8a399cf205)
# and the notebook available on
# [the authors's github](https://github.com/Christophe-pere/Time_series_RNN).
#
# We will repeat this study on starlight using the LSTM architecture to predict the observed light flux through time.
#
# Our LSTM implementation is based on this [notebook from Haiku's github repository](https://github.com/deepmind/dm-haiku/blob/master/examples/haiku_lstms.ipynb).
#
# We'll see how to use WAX-ML to ease the preparation of time series data stored in dataframes and having Nans
# before calling a "standard" deep-learning workflow.
#
# ## Disclaimer
#
# Despite the fact that this code works with real data, the results presented here should not be considered as scientific knowledge insights, to the knowledge of the authors of WAX-ML, neither the results nor the data source have been reviewed by an astrophysics pair.
#
# The purpose of this notebook is only to demonstrate how WAX-ML can be used when applying a "standard" machine learning workflow, here LSTM, to analyze time series.

# ## Download the data

register_wax_accessors()

# + tags=["parameters"]
# Parameters
STAR = "007609553"
SEQ_LEN = 64
BATCH_SIZE = 8
TRAIN_SIZE = 2**16
NUM_EPOCHS = 10
NUM_STARS = None
RECORD_FREQ = 100
TOTAL_LEN = None
TRAIN_DATE = "2016"
CACHE_DIR = Path("./cached_data/")
# -

# %%time
filename = CACHE_DIR / "kep_lightcurves.parquet"
try:
    raw_dataframe = pd.read_parquet(open(filename, "rb"))
    print(f"data read from {filename}")
except FileNotFoundError:
    # Downloading the csv file from Chrustioge Pere GitHub account
    download = requests.get(
        "https://raw.github.com/Christophe-pere/Time_series_RNN/master/kep_lightcurves.csv"
    ).content
    raw_dataframe = pd.read_csv(io.StringIO(download.decode("utf-8")))
    # set date index
    raw_dataframe.index = pd.Index(
        pd.date_range("2009-03-07", periods=len(raw_dataframe.index), freq="h"),
        name="time",
    )
    # save dataframe locally in CACHE_DIR
    CACHE_DIR.mkdir(exist_ok=True)
    raw_dataframe.to_parquet(filename)
    print(f"data saved in {filename}")


# shortening of data to speed up the execution of the notebook in the CI
if TOTAL_LEN:
    raw_dataframe = raw_dataframe.iloc[:TOTAL_LEN]


# Let's visualize the description of this dataset:

raw_dataframe.describe().T.to_xarray()


stars = raw_dataframe.columns
stars = sorted(list(set([i.split("_")[0] for i in stars])))
print(f"The number of stars available is: {len(stars)}")
print(f"star identifiers: {stars}")

dataframe = raw_dataframe[[i + "_rscl" for i in stars]].rename(
    columns=lambda c: c.replace("_rscl", "")
)
dataframe.columns.names = ["star"]
dataframe.shape

if NUM_STARS:
    columns = dataframe.columns.tolist()
    columns.remove(STAR)
    dataframe = dataframe[[STAR] + columns[: NUM_STARS - 1]]

# ## Rolling mean

# We will smooth the data by applying a rolling mean with a window of 100 periods.

# ### Count nan values
#
# But before since the dataset has some nan values, we will extract few statistics
# about the density of nan values in windows of size 100.
#
# It will be the occasion to show a usage of the `wax.modules.Buffer` module with the `format_outputs=False`
# option for the dataframe accessor `.wax.stream`.


# Let's apply the `Buffer` module to the data:

buffer, _ = dataframe.wax.stream(format_outputs=False).apply(lambda x: Buffer(100)(x))

assert isinstance(buffer, jnp.ndarray)

# Equivalently, we can use wax `unroll` function.


buffer = unroll(lambda x: Buffer(100)(x))(jax.device_put(dataframe.values))

# Let's describe the statistic of nans with pandas:

count_nan = jnp.isnan(buffer).sum(axis=1)
pd.DataFrame(onp.array(count_nan)).stack().describe().astype(int)

# ### Computing the rolling mean

# We will choose a `min_periods` of 5 in order to keep at leas 75% of the points.

# %%time
dataframe_mean, _ = dataframe.wax.stream().apply(
    lambda x: RollingMean(100, min_periods=5)(x)
)

dataframe.iloc[:, :2].plot()

# ## Forecasting with Machine Learning
#
# We need two forecast in this data, if you look with attention you'll see micro holes and big holes.


T = TypeVar("T")


class Pair(NamedTuple):
    x: T
    y: T


class TrainSplit(NamedTuple):
    train: T
    validation: T


gg.theme_set(gg.theme_bw())
warnings.filterwarnings("ignore")


plt.rcParams["figure.figsize"] = 18, 8
fig, (ax, lax) = plt.subplots(ncols=2, gridspec_kw={"width_ratios": [4, 1]})
dataframe.plot(ax=ax, title="raw data")
ax.legend(bbox_to_anchor=(0, 0, 1, 1), bbox_transform=lax.transAxes)
lax.axis("off")


plt.rcParams["figure.figsize"] = 18, 8
fig, (ax, lax) = plt.subplots(ncols=2, gridspec_kw={"width_ratios": [4, 1]})
dataframe_mean.plot(ax=ax, title="Smoothed data")
ax.legend(bbox_to_anchor=(0, 0, 1, 1), bbox_transform=lax.transAxes)
lax.axis("off")
# -

# ### Normalize data

dataframe_mean.stack().hist(bins=100, log=True)


def min_max_scaler(values: pd.DataFrame, output_format: str = "dataframe") -> Encoder:
    scaler = MinMaxScaler(feature_range=(0, 1))
    scaler.fit(values)
    index = values.index
    columns = values.columns

    def encode(dataframe: pd.DataFrame):
        nonlocal index
        nonlocal columns

        index = dataframe.index
        columns = dataframe.columns
        array_normed = scaler.transform(dataframe)

        if output_format == "dataframe":
            return pd.DataFrame(array_normed, index, columns)
        elif output_format == "jax":
            return jnp.array(array_normed)
        else:
            return array_normed

    def decode(array_scaled):

        value = scaler.inverse_transform(array_scaled)

        if output_format == "dataframe":
            return pd.DataFrame(value, index, columns)
        else:
            return value

    return Encoder(encode, decode)


# -

scaler = min_max_scaler(dataframe_mean)
dataframe_normed = scaler.encode(dataframe_mean)
assert (scaler.decode(dataframe_normed) - dataframe_mean).stack().abs().max() < 1.0e-4

dataframe_normed.stack().hist(bins=100)

# ### Prepare train / validation datasets


def split_feature_target(
    dataframe,
    look_back=SEQ_LEN,
    shuffle=True,
    stack=True,
    min_periods_ratio: float = 0.8,
    rng=None,
) -> Pair:
    def prepare_xy(data):
        buffer = Buffer(look_back + 1)(data)
        x = buffer[:-1]
        y = buffer[-1]
        return x, y

    def prepare_xy(data):
        y = Buffer(look_back)(data)
        x = Lag(1)(y)
        return x, y

    x, y = unroll(prepare_xy)(jax.device_put(dataframe.values))

    if shuffle:
        if rng is None:
            rng = jax.random.PRNGKey(42)

        B = x.shape[0]
        idx = jnp.arange(B)
        idx = jax.random.shuffle(rng, idx)

        x = x[idx]
        y = y[idx]

    if stack:
        B, T, F = x.shape
        x = x.transpose(1, 0, 2).reshape(T, B * F, 1).transpose(1, 0, 2)
        y = y.transpose(1, 0, 2).reshape(T, B * F, 1).transpose(1, 0, 2)

    if min_periods_ratio:
        T = x.shape[1]
        count_nan = jnp.isnan(x).sum(axis=1)
        mask = count_nan < min_periods_ratio * T
        idx = jnp.where(mask)
        x = x[idx[0]]
        y = y[idx[0]]

    # round Batch size to a power of to
    B = x.shape[0]
    B_round = int(2 ** jnp.floor(jnp.log2(B)))
    print(f"{B} batches rounded to {B_round} batches.")
    x = x[:B_round]
    y = y[:B_round]

    # fillnan by zeros
    x, y = hk.testing.transform_and_run(lambda x: FillNanInf()(x))((x, y))

    return Pair(x, y)


# +
# split_feature_target(dataframe)
# -


def split_train_validation(
    dataframe, train_size, look_back, scaler: Optional[Callable] = None
) -> TrainSplit:

    # prepare scaler
    train_df = dataframe.iloc[:train_size]

    if scaler:
        scaler = scaler(train_df)

    # prepare train data
    if scaler:
        train_df = scaler.encode(train_df)

    train_xy = split_feature_target(train_df, look_back)

    # prepare validation data
    valid_size = len(dataframe) - train_size
    valid_size = int(2 ** jnp.floor(jnp.log2(valid_size)))

    valid_end = int(train_size + valid_size)
    valid_df = dataframe.iloc[train_size:valid_end]

    if scaler:
        valid_df = scaler.encode(valid_df)

    valid_xy = split_feature_target(valid_df, look_back)

    return TrainSplit(train_xy, valid_xy)


TRAIN_SIZE

print(f"Look at star: {STAR}")
train, valid = split_train_validation(dataframe_normed[[STAR]], TRAIN_SIZE, SEQ_LEN)

train[0].shape, train[1].shape, valid[0].shape, valid[1].shape

# TRAIN_SIZE, VALID_SIZE = len(train.x), len(valid.x)
print(
    f"effective train_size = {len(train.x)}, " f"effective valid size= {len(valid.x)}"
)


# Plot an observation/target pair.
rng = jax.random.PRNGKey(42)
batch_plot = jax.random.choice(rng, len(train[0]))
df = pd.DataFrame(
    {"x": train.x[batch_plot, :, 0], "y": train.y[batch_plot, :, 0]}
).reset_index()
df = pd.melt(df, id_vars=["index"], value_vars=["x", "y"])
plot = (
    gg.ggplot(df)
    + gg.aes(x="index", y="value", color="variable")
    + gg.geom_line()
    + gg.scales.scale_y_log10()
)
_ = plot.draw()


# ### Dataset iterator


class Dataset:
    """An iterator over a numpy array, revealing batch_size elements at a time."""

    def __init__(self, xy: Pair, batch_size: int):
        self._x, self._y = xy
        self._batch_size = batch_size
        self._length = self._x.shape[0]
        self._idx = 0
        if self._length % batch_size != 0:
            msg = "dataset size {} must be divisible by batch_size {}."
            raise ValueError(msg.format(self._length, batch_size))

    def __next__(self) -> Pair:
        start = self._idx
        end = start + self._batch_size
        x, y = self._x[start:end], self._y[start:end]
        if end >= self._length:
            print(f"End of the data set (size={end}). Return to the beginning.")
            end = end % self._length
            assert end == 0  # Guaranteed by ctor assertion.
        self._idx = end
        return Pair(x, y)


# + [markdown] colab_type="text" id="LZGw5Jdvjmqh"
# ### Training an LSTM
#
# To train the LSTM, we define a Haiku function which unrolls the LSTM over the input sequence, generating predictions for all output values. The LSTM always starts with its initial state at the start of the sequence.
#
# The Haiku function is then transformed into a pure function through `hk.transform`, and is trained with Adam on an L2 prediction loss.


# + colab={} colab_type="code" id="nacnTj5ejIK5"
def unroll_net(seqs: jnp.ndarray):
    """Unrolls an LSTM over seqs, mapping each output to a scalar."""
    # seqs is [T, B, F].
    core = hk.LSTM(32)
    batch_size = seqs.shape[0]
    outs, state = hk.dynamic_unroll(
        core, seqs, core.initial_state(batch_size), time_major=False
    )
    # We could include this Linear as part of the recurrent core!
    # However, it's more efficient on modern accelerators to run the linear once
    # over the entire sequence than once per sequence element.
    return hk.BatchApply(hk.Linear(1))(outs), state


# + colab={} colab_type="code" id="nacnTj5ejIK5"
model = jit_init_apply(hk.transform(unroll_net))


# +
@jax.jit
def loss(pred, y):
    return jnp.mean(jnp.square(pred - y))


def model_with_loss(x, y):
    pred, _ = unroll_net(x)
    return loss(pred, y)


# + colab={} colab_type="code" id="nacnTj5ejIK5"
class TrainState(NamedTuple):
    step: int
    params: Any
    opt_state: Any
    rng: jnp.ndarray
    loss: float


def train_model(
    model_with_loss: Callable,
    train_ds: Dataset,
    valid_ds: Dataset,
    max_iterations: int = -1,
    rng=None,
    record_freq=100,
) -> hk.Params:
    """Initializes and trains a model on train_ds, returning the final params."""
    opt = optax.adam(1e-3)
    model_with_loss = jit_init_apply(hk.transform(model_with_loss))

    @jax.jit
    def update(train_state, x, y):
        step, params, opt_state, rng, _ = train_state
        if rng is not None:
            (rng,) = jax.random.split(rng, 1)
        l, grads = jax.value_and_grad(model_with_loss.apply)(params, rng, x, y)
        grads, opt_state = opt.update(grads, opt_state)
        params = optax.apply_updates(params, grads)
        return TrainState(step + 1, params, opt_state, rng, l)

    # Initialize state.
    def init():
        x, y = next(train_ds)
        params = model_with_loss.init(rng, x, y)
        opt_state = opt.init(params)
        return TrainState(0, params, opt_state, rng, jnp.inf)

    def _format_results(records):
        records = {key: jnp.stack(l) for key, l in records.items()}
        return records

    records = defaultdict(list)
    train_state = init()
    with tqdm(total=max_iterations if max_iterations > 0 else None) as pbar:
        while True:
            try:
                x, y = next(train_ds)
            except StopIteration:
                return train_state, _format_results(records)

            train_state = update(train_state, x, y)
            if train_state.step % record_freq == 0:
                x, y = next(valid_ds)
                if rng is not None:
                    (rng,) = jax.random.split(rng, 1)
                valid_loss = model_with_loss.apply(train_state.params, rng, x, y)
                records["step"].append(train_state.step)
                records["valid_loss"].append(valid_loss)
                records["train_loss"].append(train_state.loss)

            pbar.update()
            if max_iterations > 0 and train_state.step >= max_iterations:
                return train_state, _format_results(records)


# + colab={} colab_type="code" id="AssgDctokbl5"
# %%time
train, valid = split_train_validation(dataframe_normed[[STAR]], TRAIN_SIZE, SEQ_LEN)
train_ds = Dataset(train, BATCH_SIZE)
valid_ds = Dataset(valid, BATCH_SIZE)


train_state, records = train_model(
    model_with_loss,
    train_ds,
    valid_ds,
    len(train.x) // BATCH_SIZE * NUM_EPOCHS,
    rng=jax.random.PRNGKey(42),
    record_freq=RECORD_FREQ,
)

# +
# train_state.params
# -

# Plot losses
losses = pd.DataFrame(records)
df = pd.melt(losses, id_vars=["step"], value_vars=["train_loss", "valid_loss"])
plot = (
    gg.ggplot(df)
    + gg.aes(x="step", y="value", color="variable")
    + gg.geom_line()
    + gg.scales.scale_y_log10()
)
_ = plot.draw()


# + [markdown] colab_type="text" id="yr7jrOL3ki-b"
# ### Sampling
#
# The point of training models is so that they can make predictions! How can we generate predictions with the trained model?
#
# If we're allowed to feed in the ground truth, we can just run the original model's `apply` function.

# + colab={} colab_type="code" id="f2qETEqXLT1N"
def plot_samples(truth: np.ndarray, prediction: np.ndarray) -> gg.ggplot:
    assert truth.shape == prediction.shape
    df = pd.DataFrame(
        {"truth": truth.squeeze(), "predicted": prediction.squeeze()}
    ).reset_index()
    df = pd.melt(df, id_vars=["index"], value_vars=["truth", "predicted"])
    plot = (
        gg.ggplot(df) + gg.aes(x="index", y="value", color="variable") + gg.geom_line()
    )
    return plot


# + colab={} colab_type="code" id="KOuK1egilGD0"
# Grab a sample from the validation set.
sample_x, sample_y = next(valid_ds)
sample_x = sample_x[:1]  # Shrink to batch-size 1.
sample_y = sample_y[:1]

# Generate a prediction, feeding in ground truth at each point as input.
predicted, _ = model.apply(train_state.params, None, sample_x)

plot = plot_samples(sample_y, predicted)
plot.draw()
del sample_x, predicted


# -

# ### Run autoregressively

# + [markdown] colab_type="text" id="tDyGshz_lwrM"
# If we can't feed in the ground truth (because we don't have it), we can also run the model autoregressively.

# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
def autoregressive_predict(
    trained_params: hk.Params,
    context: jnp.ndarray,
    seq_len: int,
    pbar=False,
):
    """Given a context, autoregressively generate the rest of a sine wave."""

    ar_outs = []
    context = jax.device_put(context)
    times = onp.arange(seq_len - context.shape[1] + 1)
    if pbar:
        times = tqdm(times)
    for _ in times:
        full_context = jnp.concatenate([context] + ar_outs, axis=1)

        outs, _ = model.apply(trained_params, None, full_context)
        # Append the newest prediction to ar_outs.
        ar_outs.append(outs[:, -1:, :])
    # Return the final full prediction.
    return outs


# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
sample_x, sample_y = next(valid_ds)
sample_x = sample_x[:1]  # Shrink to batch-size 1.
sample_y = sample_y[:1]  # Shrink to batch-size 1.


context_length = SEQ_LEN // 8
print(f"context_length = {context_length}")
# Cut the batch-size 1 context from the start of the sequence.
context = sample_x[:, :context_length]

# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
# %%time
# We can reuse params we got from training for inference - as long as the
# declaration order is the same.
predicted = autoregressive_predict(train_state.params, context, SEQ_LEN, pbar=True)
# -

sample_y.shape, predicted.shape

# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
plot = plot_samples(sample_y, predicted)
plot += gg.geom_vline(xintercept=context.shape[1], linetype="dashed")
_ = plot.draw()


# + [markdown] colab_type="text" id="qGkr2gf2oALo"
# #### Sharing parameters with a different function.
#
# Unfortunately, this is a bit slow - we're doing O(N^2) computation for a sequence of length N.
#
# It'd be better if we could do the autoregressive sampling all at once - but we need to write a new Haiku function for that.
#
# We're in luck - if the Haiku module names match, the same parameters can be used for multiple Haiku functions.
#
# This can be achieved through a combination of two techniques:
#
# 1. If we manually give a unique name to a module, we can ensure that the parameters are directed to the right places.
# 2. If modules are instantiated in the same order, they'll have the same names in different functions.
#
# Here, we rely on method #2 to create a fast autoregressive prediction.

# + colab_type="text" id="qGkr2gf2oALo"
@hk.transform
def fast_autoregressive_predict_fn(context, seq_len):
    """Given a context, autoregressively generate the rest of a sine wave."""
    core = hk.LSTM(32)
    dense = hk.Linear(1)
    state = core.initial_state(context.shape[0])
    # Unroll over the context using `hk.dynamic_unroll`.
    # As before, we `hk.BatchApply` the Linear for efficiency.
    context_outs, state = hk.dynamic_unroll(
        core,
        context,
        state,
        time_major=False,
    )
    context_outs = hk.BatchApply(dense)(context_outs)

    # Now, unroll one step at a time using the running recurrent state.
    ar_outs = []
    x = context_outs[:, -1, :]
    times = range(seq_len - context.shape[1])
    for _ in times:
        x, state = core(x, state)
        x = dense(x)
        ar_outs.append(x)
    ar_outs = jnp.stack(ar_outs)
    ar_outs = ar_outs.transpose(1, 0, 2)
    return jnp.concatenate([context_outs, ar_outs], axis=1)


fast_autoregressive_predict = jax.jit(
    fast_autoregressive_predict_fn.apply, static_argnums=(3,)
)

# + colab={} colab_type="code" id="WdKcHr6_n_ba"
# %%time
# Reuse the same context from the previous cell.
predicted = fast_autoregressive_predict(train_state.params, None, context, SEQ_LEN)


# + colab={} colab_type="code" id="WdKcHr6_n_ba"
# The plots should be equivalent!
plot = plot_samples(sample_y, predicted)
plot += gg.geom_vline(xintercept=context.shape[1], linetype="dashed")
_ = plot.draw()
# -


# # Sample trajectories

# + colab={} colab_type="code" id="WdKcHr6_n_ba"
sample_x, sample_y = next(valid_ds)
sample_x = sample_x[:1]  # Shrink to batch-size 1.
sample_y = sample_y[:1]  # Shrink to batch-size 1.


context_length = SEQ_LEN // 8
print(f"context_length = {context_length}")
# Cut the batch-size 1 context from the start of the sequence.
context = sample_x[:, :context_length]

# Reuse the same context from the previous cell.
predicted = fast_autoregressive_predict(train_state.params, None, context, SEQ_LEN)

# The plots should be equivalent!
plot = plot_samples(sample_y, predicted)
plot += gg.geom_vline(xintercept=context.shape[1], linetype="dashed")
_ = plot.draw()
# -


# ## timeit

# + colab={} colab_type="code" id="9S0tkPXGrU3a"
# %timeit autoregressive_predict(train_state.params, context, SEQ_LEN)
# %timeit fast_autoregressive_predict(train_state.params, None, context, SEQ_LEN)
# -
# ## Train all stars


# ### Training


def split_train_validation_date(dataframe, date, look_back) -> TrainSplit:
    train_size = len(dataframe.loc[:date])
    return split_train_validation(dataframe, train_size, look_back)


# %%time
train, valid = split_train_validation_date(dataframe_normed, TRAIN_DATE, SEQ_LEN)
print(f"effective train size = {train[0].shape[1]}")

train[0].shape, train[1].shape, valid[0].shape, valid[1].shape

train_ds = Dataset(train, BATCH_SIZE)
valid_ds = Dataset(valid, BATCH_SIZE)
# del train, valid  # Don't leak temporaries.

# %%time
train_state, records = train_model(
    model_with_loss,
    train_ds,
    valid_ds,
    len(train.x) // BATCH_SIZE * 1,
    jax.random.PRNGKey(42),
    record_freq=RECORD_FREQ,
)

# Plot losses
losses = pd.DataFrame(records)
df = pd.melt(losses, id_vars=["step"], value_vars=["train_loss", "valid_loss"])
plot = (
    gg.ggplot(df)
    + gg.aes(x="step", y="value", color="variable")
    + gg.geom_line()
    + gg.scales.scale_y_log10()
)
_ = plot.draw()

# ### Sampling

# +
# Grab a sample from the validation set.
sample_x, sample_y = next(valid_ds)
sample_x = sample_x[:1]  # Shrink to batch-size 1.
sample_y = sample_y[:1]  # Shrink to batch-size 1.


# Generate a prediction, feeding in ground truth at each point as input.
predicted, _ = model.apply(train_state.params, None, sample_x)

plot = plot_samples(sample_y, predicted)
_ = plot.draw()
# -

# ### Run autoregressively

# +
# %%time
sample_x, sample_y = next(valid_ds)
sample_x = sample_x[:1]  # Shrink to batch-size 1.
sample_y = sample_y[:1]  # Shrink to batch-size 1.


context_length = SEQ_LEN // 8
# Cut the batch-size 1 context from the start of the sequence.
context = sample_x[:, :context_length]

# Reuse the same context from the previous cell.
predicted = fast_autoregressive_predict(train_state.params, None, context, SEQ_LEN)

# The plots should be equivalent!
plot = plot_samples(sample_y, predicted)
plot += gg.geom_vline(xintercept=len(context), linetype="dashed")
_ = plot.draw()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

import inspect

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# -- Options for LaTeX output ------------------------------------------------
from typing import Dict, Set

sys.path.insert(0, os.path.abspath(".."))


gh_url = ""

# -- Project information -----------------------------------------------------

project = "WAX-ML"
copyright = "2021, WAX-ML Authors"
author = "The WAX-ML Authors"

# The short X.Y version
version = ""
# The full version, including alpha/beta/rc tags
release = ""


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = '2.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'matplotlib.sphinxext.plot_directive',
    'sphinx_autodoc_typehints',
    'myst_nb',
    ]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'jax': ('https://jax.readthedocs.io/en/stable/', None),
    "haiku" : ('https://dm-haiku.readthedocs.io/en/latest/', None),
}

suppress_warnings = [
    'ref.citation',  # Many duplicated citations in numpy/scipy docstrings.
    'ref.footnote',  # Many unreferenced footnotes in numpy/scipy docstrings
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# The suffix(es) of source filenames.
# Note: important to list ipynb before md here: we have both md and ipynb
# copies of each notebook, and myst will choose which to convert based on
# the order in the source_suffix list. Notebooks which are not executed have
# outputs stored in ipynb but not in md, so we must convert the ipynb.
source_suffix = ['.rst', '.ipynb', '.md']



# The master toctree document.
main_doc = 'index'


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    # Sometimes sphinx reads its own outputs as inputs!
    'build/html',
    'build/jupyter_execute',
    'notebooks/README.md',
    'README.md',
    'CHANGELOG.md',
    # Ignore markdown source for notebooks; myst-nb builds from the ipynb
    # These are kept in sync using the jupytext pre-commit hook.
    'notebooks/*.md',
    'notebooks/*.py',
]


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None
# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "friendly"


autosummary_generate = True
napolean_use_rtype = False


# mathjax_config = {
#     'TeX': {'equationNumbers': {'autoNumber': 'AMS', 'useLabelIds': True}},
# }

# Additional files needed for generating LaTeX/PDF output:
# latex_additional_files = ['references.bib']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'logo_only': True,
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.

html_logo = '_static/wax_logo_250px.png'
html_favicon = '_static/favicon.png'


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}
# html_css_files = ["css/custom.css"]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for myst ----------------------------------------------
jupyter_execute_notebooks = "force"
execution_allow_errors = False
execution_fail_on_error = True  # Requires https://github.com/executablebooks/MyST-NB/pull/296

# Notebook cell execution timeout; defaults to 30.
execution_timeout = 100

# List of patterns, relative to source directory, that match notebook
# files that will not be executed.
import glob
execution_excludepatterns = list(glob.glob("notebooks/*"))

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "WAXdoc"


latex_elements: Dict = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).

latex_documents = [
    (main_doc, 'WAX-ML.tex', 'WAX-ML Documentation',
     'The WAX-ML authors', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (main_doc, "wax", "WAX-ML Documentation", [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        main_doc,
        "WAX-ML",
        "WAX-ML Documentation",
        author,
        "WAX-ML",
        "Oneline description of project."
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

autodoc_member_order = "bysource"


# -- Extension configuration -------------------------------------------------

# Tell sphinx-autodoc-typehints to generate stub parameter annotations including
# types, even if the parameters aren't explicitly documented.
always_document_param_types = True

# -- Options for intersphinx extension ---------------------------------------


# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for numpydocs extension -----------------------------------------
# https://numpydoc.readthedocs.io/en/latest/install.html

# Whether to produce plot:: directives for Examples sections that contain
# import matplotlib or from matplotlib import.
numpydoc_use_plots = True

# Whether to show all members of a class in the Methods and Attributes sections
# automatically. True by default.
numpydoc_show_class_members = True

#  Whether to show all inherited members of a class in the Methods and
#  Attributes sections automatically. If it’s false, inherited members won’t
#  shown. True by default.
numpydoc_show_inherited_class_members = True

#  Whether to create a Sphinx table of contents for the lists of class methods
#  and attributes. If a table of contents is made, Sphinx expects each entry to
#  have a separate page. True by default.
numpydoc_class_members_toctree = False

# A regular expression matching citations which should be mangled to avoid
# conflicts due to duplication across the documentation. Defaults to [\w-]+.
numpydoc_citation_re = r"[\w-]+"

# Until version 0.8, parameter definitions were shown as blockquotes, rather
# than in a definition list. If your styling requires blockquotes, switch this
# config option to True. This option will be removed in version 0.10.
numpydoc_use_blockquotes = False

# Whether to format the Attributes section of a class page in the same way as
# the Parameter section. If it's False, the Attributes section will be
# formatted as the Methods section using an autosummary table. True by default.
numpydoc_attributes_as_param_list = False

# Whether to create cross-references for the parameter types in the Parameters,
# Other Parameters, Returns and Yields sections of the docstring. False by
# default.
numpydoc_xref_param_type = False

#  Mappings to fully qualified paths (or correct ReST references) for the
#  aliases/shortcuts used when specifying the types of parameters. The keys
#  should not have any spaces. Together with the intersphinx extension, you can
#  map to links in any documentation. The default is an empty dict.  This
#  option depends on the numpydoc_xref_param_type option being True.
numpydoc_xref_aliases: Dict = {}

#  Words not to cross-reference. Most likely, these are common words used in
#  parameter type descriptions that may be confused for classes of the same
#  name. For example: {'type', 'optional', 'default'}. The default is an empty
#  set.

numpydoc_xref_ignore: Set = set([])

#  Deprecated since version edit: your HTML template instead Whether to insert
#  an edit link after docstrings.
numpydoc_edit_link: bool



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define `Stream` object used to synchronize in-memory data streams and
unroll data transformations on it.
"""
from collections import defaultdict, deque
from dataclasses import dataclass, field
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Hashable,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

import jax
import numpy as onp
import pandas as pd
import xarray as xr
from jax import numpy as jnp
from jax.tree_util import tree_flatten, tree_leaves, tree_map, tree_unflatten
from tqdm.auto import tqdm

from wax.encode import (
    Encoder,
    datetime64_encoder,
    encode_dataset,
    floor_datetime,
    string_encoder,
)
from wax.unroll import UnrollTransformedWithState, unroll_transform_with_state
from wax.utils import get_unique_dtype

# DTypeLike = TypeVar("DTypeLike")
DTypeLike = type

EncoderMapping = Union[
    Dict[Hashable, Callable[[Any], Encoder]], Dict[Hashable, Encoder]
]


DEFAULT_TYPES_ENCODERS = cast(
    EncoderMapping,
    {
        onp.dtype(str).type: string_encoder,
        onp.datetime64: datetime64_encoder,
    },
)
DTYPE_INIT_VALUES = {
    jnp.float32: jnp.nan,
    jnp.float64: jnp.nan,
    jnp.int64: -99999,
    jnp.int32: -99999,
    jnp.uint64: 0,
    jnp.uint32: 0,
    # numpy
    onp.float32: onp.nan,
    onp.float64: onp.nan,
    onp.int64: -99999,
    onp.int32: -99999,
    onp.uint64: 0,
    onp.uint32: 0,
    bool: False,
    onp.bool_: False,
}


def timestamp_is_before_local(
    embed: Callable,
    local_timestamp: onp.ndarray,
    secondary_timestamp: onp.ndarray,
) -> bool:
    """Says if a secondary timestamp is before the local timestamp. It uses an `embed` callable
    to convert the `local_time` into the `other_time` referential and thus allowing
    comparison.

    Args:
        embed : callable used to embed `local_time` in `other_time`.
        local_timestamp : local timestamp
        secondary_timestamp : other timestamp

    Returns:
        true if first timestamp is before the second,
        false otherwise.
    """
    local_time_embedded = embed(local_timestamp)
    return secondary_timestamp < local_time_embedded


class StreamObservation(NamedTuple):
    time: Any
    value: Any


class DatasetSchema(NamedTuple):
    coords: xr.core.coordinates.DatasetCoordinates
    encoders: Dict[Hashable, Encoder]


def _is_verbose(verbose, time_dim):
    if isinstance(verbose, bool):
        return verbose

    if isinstance(verbose, (list, tuple)):
        if time_dim in verbose:
            return True
    return False


class GeneratorState(NamedTuple):
    output_values: Any
    stream_obs: Any


def get_time_dataset(dataset):
    """ """
    time_coords = get_dataset_time_coords(dataset)
    time_dataset = xr.Dataset()
    for time_dim in time_coords:
        # time_dataset[time_dim + "_index"] = dataset[time_dim]
        time_dataset[time_dim + "_index"] = xr.DataArray(
            onp.arange(len(dataset[time_dim])),
            dims=(time_dim,),
            coords={time_dim: dataset[time_dim].values},
        )

    return time_dataset


def get_dataset_index_from_stream_index(
    stream_index: Mapping[str, "StreamObservation"]
) -> xr.Dataset:
    """Convert stream_data index to dataset index.

     Args:
         stream_index : mapping of StreamObservation items with numpy index in "value" attribute.
    Returns:
         dataset with "step" coordinate and numpy index as variables.
            Buffered streams have an additional coordinate 'time_dim'+ '_buffer_index'
            construct buffers of data.
    """
    dataset = xr.Dataset()
    for time_dim, idx in stream_index.items():
        assert isinstance(idx, StreamObservation)
        idx_value = idx.value[time_dim + "_index"]
        if idx_value.ndim == 2:
            dataset[time_dim] = xr.DataArray(
                idx_value, dims=("step", time_dim + "_buffer_index")
            )
        else:
            assert idx_value.ndim == 1
            dataset[time_dim] = xr.DataArray(idx_value, dims="step")

    return dataset


def _get_unique_time_dim(dataset):
    time_coords = {
        dim for dim, vals in dataset.coords.items() if vals.dtype.type is onp.datetime64
    }
    assert len(time_coords) == 1
    return time_coords.pop()


def get_dataset_time_coords(dataset):
    time_coords = {
        dim for dim, vals in dataset.coords.items() if vals.dtype.type is onp.datetime64
    }
    assert len(time_coords), "No time coordinate found"
    return time_coords


def split_dataset_from_time_dims(dataset):
    time_coords = get_dataset_time_coords(dataset)

    datasets = defaultdict(xr.Dataset)

    # create empty dataset with time coords
    for time_dim in time_coords:
        datasets[time_dim][time_dim] = dataset[time_dim]

    # fill datasets with the input dataset variables
    for var_dim, var in dataset.items():
        var_time_dims = time_coords.intersection(set(var.dims))
        if var_time_dims:
            assert len(var_time_dims) == 1, (
                f"variable {var_dim} has multiple time dimensions {var_time_dims}."
                "It should contain only one."
            )
            var_time_dim = var_time_dims.pop()
            datasets[var_time_dim][var_dim] = var

    return datasets


def get_dataset_index(
    dataset: xr.Dataset, time_dataset_index: xr.Dataset
) -> xr.Dataset:
    """Construct dataset of index from a dataset and a time_dataset_index.

    Args:
        dataset : original dataset
        time_datset_index : dataset of index from different timelines
            constructed with the original dataset.

    Returns:
        Dataset of index.
    """
    dataset_index = xr.Dataset()
    n_steps = len(next(iter(time_dataset_index.values())))
    for dim, var in dataset.items():
        if set(var.dims).intersection(time_dataset_index.keys()):
            time_dim = var.dims[0]
            assert time_dim in time_dataset_index, (
                f"'{time_dim}' does not seems to be a time "
                f"dimensions in {time_dataset_index.keys()}. "
                "For the moment, only time dimension as first dim is supported."
            )
            dataset_index[dim] = time_dataset_index[time_dim]
        else:
            if not onp.shape(var):
                dataset_index[dim] = xr.DataArray(onp.arange(n_steps), dims=("step",))
            else:
                values_atleast_1d = onp.atleast_1d(var.values)
                # grid = onp.indices(values_atleast_1d.shape)
                flat_idx = onp.arange(len(values_atleast_1d.ravel()))
                dataset_index[dim] = xr.DataArray(
                    onp.outer(onp.arange(n_steps), flat_idx),
                    dims=("step", cast(str, dim) + "_flat_idx"),
                )
    return dataset_index


def tree_buffer_transform(maxlen, fill_value, init_value):
    init_value, treedef = tree_flatten(init_value)
    fill_value = tree_leaves(fill_value)
    tree_len = len(init_value)
    assert len(fill_value) == tree_len
    buffer = [deque(maxlen=maxlen) for x in init_value]

    def transform(data):
        data = tree_leaves(data)
        assert len(data) == tree_len
        output = []
        for b, x, fv in zip(buffer, data, fill_value):
            b.append(x)
            output.append(deque_to_list(b, fv))
        return tree_unflatten(treedef, output)

    return transform


def deque_to_list(buffer: deque, fill_value: Any):
    """Convert buffer to list of length maxlen and replace missing
    observations with fill_value.

    Args:
        buffer_list : deque
        fill_value : fill_value to use
    """
    maxlen = cast(int, buffer.maxlen)
    buffer_list = list(buffer)

    x = buffer_list[0]
    if x is pd.NaT or x.dtype in [
        onp.dtype("datetime64[ns]"),
        onp.dtype("<M8[ns]"),
        onp.dtype("<M8[us]"),
    ]:
        if len(buffer_list) < maxlen:
            buffer_list = [pd.NaT] * (maxlen - len(buffer_list)) + buffer_list
        # return a deque to avoid flattening if later processing (like in stream_unroll)
        return deque(buffer_list)
    else:
        if len(buffer_list) < maxlen:
            buffer_list = [onp.full_like(buffer_list[0], fill_value)] * (
                maxlen - len(buffer_list)
            ) + buffer_list
        return onp.stack(buffer_list)


def unroll_stream(
    stream: Generator, skip_first: bool = False, pbar: Union[bool, Sequence[Any]] = True
) -> Any:
    """Unroll a generator by collecting its outputs in a list.

    Args:
        stream : generator
        skip_first : if true, skip first observation
        pbar : if true, activate progressbar.

    Yield:
        Nested data structure with list of collected outputs as leaves.
    """
    # init
    obs = next(stream)

    obs_flat, treedef = tree_flatten(obs)
    num_leaves = len(obs_flat)

    # stream_scan
    def _init_outputs():
        if skip_first:
            return [[]] * num_leaves
        else:
            return list(map(lambda x: [x], obs_flat))

    outputs = _init_outputs()

    if pbar:
        stream = tqdm(stream, desc="stream_unroll")

    for obs in stream:
        obs_flat = tree_leaves(obs)
        assert len(obs_flat) == num_leaves
        for y, x in zip(outputs, obs_flat):
            y.append(x)

    # stack outputs
    for i in range(num_leaves):
        outputs[i] = onp.stack(outputs[i])

    # transpose outputs
    return tree_unflatten(treedef, outputs)


def dataset_to_numpy(dataset) -> Dict[str, jnp.ndarray]:
    """Convert a xarray Dataset into a dict of numpy arrays."""
    dataset_numpy = {dim: var.values for dim, var in dataset.items()}
    return dataset_numpy


def np_select_item(np_dataset: Dict[str, Any], i: int, trace: bool = False):
    """Select item of a dataset.

    Args:
        np_dataset : dict of numpy arrays
        i : element position to access.
        trace : if true, replace the actual value in the output by its position.
    """
    if trace:
        # return only indices
        return {dim: onp.array(i) for dim, var in np_dataset.items()}
    else:
        # return real data
        return {dim: var[i] for dim, var in np_dataset.items()}


def access_data(step: int, x: Any, idx: Any):
    """Access to an array at a given step using a predifined access index.

    Args:
        step : step of the observation.
        x : data to access
        idx : predifined access index.
    """
    step_idx = idx[step]
    # output = x[step_idx]
    output = jnp.where(
        step_idx < 0,
        jnp.full_like(jnp.atleast_1d(x)[step_idx], DTYPE_INIT_VALUES[x.dtype.type]),
        jnp.atleast_1d(x)[step_idx],
    )
    return output


def tree_access_data(data, index, step):
    """Access to data from given indices.

    Args:
        data : nested data structure with arrays leaves.
        index : nested data structure with index leaves permitting
            to access the data at a given step.
        step : step on which to access.
    """
    return tree_map(partial(access_data, step), data, index)


@dataclass(frozen=True)
class Stream:
    """Stream object used to synchronize in-memory data streams and
    unroll data transformations on it.

    We implement a synchronization
    mechanism between different data streams. Using the terminology of Henri Poincaré,
    we introduce the notion of "local time" to unravel the stream in which
    the user wants to apply transformations. We call the other streams "secondary streams".
    They can work at different frequencies, lower or higher.  The data from these secondary
    streams will be represented in the "local time" either with the use of a
    forward filling mechanism for lower frequencies or a buffering mechanism
    for higher frequencies.

    We implement a "data tracing" mechanism to optimize access to out-of-sync streams.
    This mechanism works on in-memory data.  We perform the first pass on the data,
    without actually accessing it, and determine the indices necessary to
    later access the data. Doing so we are vigilant to not let any "future"
    information pass through and thus guaranty a data processing that respects causality.


    The buffering mechanism used in the case of higher frequencies works with a fixed
    buffer size (see the WAX-ML module
    [`wax.modules.Buffer`](https://wax-ml.readthedocs.io/en/latest/_autosummary/wax.modules.buffer.html#module-wax.modules.buffer))  # noqa
    which allows us to use JAX / XLA optimizations and have efficient processing.

    Args:
        local_time : dimension along which we want to iterate.
        freqs : mapping of frequencies used to embed local_time in lower frequency streams.
        ffills : mapping of bool to ffill secondary streams.
        buffer_maxlen : mapping of int describing buffer size for secondary streams
            during one iteration of the local stream_data.
            This is needed to guaranty a known in advance data indexing scheme
            compatible with XLA compilation.
            If not specified, only the last observation of secondary streams is conserved.
        buffer_dtype_init_values : mapping of initial values to consider per dtype for
            Buffers initialization.
        verbose:  if true activate prints in the data tracing process.
        tensor_type : tensor type used for convertion of the data.
        trace : if true, perform data tracing returning only indexes instead of actual data.
        format_outputs : if true, format outputs in numpy array, otherwise return outputs
            in raw format (Jax tensors).
        return_state: if true, return state, otherwise only return unrolled outputs.

    *Terminology*: We use the name `local_time` to describe the time referential,
    taken from those identified in the input data, in which the user wants to work at the end.
    This terminology is the same as that used by Henri Poincaré in his discussion of
    the synchronization problem.
    See https://en.wikipedia.org/wiki/Einstein_synchronisation for more details.
    """

    local_time: str = ""
    freqs: Dict[str, str] = field(default_factory=dict)
    ffills: Dict[str, bool] = field(default_factory=dict)
    buffer_maxlen: Dict[str, int] = field(default_factory=dict)
    buffer_dtype_init_values: Dict[DTypeLike, Any] = field(
        default_factory=lambda: DTYPE_INIT_VALUES
    )
    verbose: bool = False
    pbar: Union[bool, Sequence] = False
    tensor_type: str = "jax"
    trace: bool = True
    format_outputs: bool = True
    return_state: bool = True

    @staticmethod
    def get_dataset_schema(
        xdata: xr.Dataset,
        types_encoders: Optional[EncoderMapping] = None,
        check: bool = False,
    ) -> DatasetSchema:
        """Determine data schema of a Dataset.

        Args:
            xdata : xarray Dataset
            types_encoders : Mapping of used for different types.
            check : if true, check encoders by testing a round-trip
                conversion on detected types.

        Returns:
            Dataset schema
        """
        if types_encoders is None:
            types_encoders = DEFAULT_TYPES_ENCODERS
        encoders = {}
        np_data = {}
        np_data.update({dim: var.values for dim, var in xdata.items()})
        np_data.update({dim: var.values for dim, var in xdata.coords.items()})
        for dim, np_var in np_data.items():
            if np_var.dtype.type in types_encoders.keys():
                encoder = types_encoders[np_var.dtype.type]
                if callable(encoder):
                    encoder = encoder(np_var)
                encoders[dim] = encoder
                if check:
                    codes = encoder.encode(np_var)
                    assert (encoder.decode(codes) == np_var).ravel().all()
        schema = DatasetSchema(coords=xdata.coords, encoders=encoders)
        return schema

    def get_encoders(self, dataset):
        schema = self.get_dataset_schema(dataset)
        return schema.encoders

    def prepare(
        self,
        dataset: xr.Dataset,
        module: Callable,
        encoders: Optional[EncoderMapping] = None,
        skip_first=False,
    ) -> Tuple[UnrollTransformedWithState, jnp.ndarray]:
        """Prepare a function that wraps the input function with the actual data and indices
         in a pair of pure functions (TransformedWithState tuple).

        Args:
            dataset : dataset on which the transformation is applied
            module : callable being able to be transformed with Haiku transform_with_state.
            encoders : encoders used to encode numpy dtypes which are not supported by Jax.
            skip_first : argument passed to unroll_transform_with_state
        Returns:
            transform_dataset: transformed function ready to process in-memory data in local time.
            xs : range of steps in local time.
        """
        encoders = encoders if encoders is not None else self.get_encoders(dataset)

        # Prepare np_data and np_index
        np_data, np_index, xs = self.trace_dataset(dataset)
        # encode values
        np_data = encode_dataset(encoders, np_data)

        # convert to jax
        if self.tensor_type == "jax" and not jax.config.jax_enable_x64:
            # explicitly convert in onp.float32 and int32
            # berfore jax conversion to avoid jax warnings.
            def np_convert(x):
                if x.dtype == onp.float64:
                    return x.astype(onp.float32)
                if x.dtype == onp.int64:
                    return x.astype(onp.int32)
                return x

            np_data, np_index, xs = tree_map(np_convert, (np_data, np_index, xs))
        np_data, np_index, xs = jax.device_put((np_data, np_index, xs))

        # @jit_init_apply
        @partial(unroll_transform_with_state, skip_first=skip_first)
        def transform_dataset(step):
            dataset = partial(tree_access_data, np_data, np_index)(step)
            return module(dataset)

        return transform_dataset, xs

    def get_local_time(self, time_datasets):
        if not self.local_time:
            if not len(time_datasets) == 1:
                raise ValueError(
                    "local_time must be specified since multiple time index have been detected.",
                    f"It should be in {time_datasets.keys()}",
                )
            local_time = next(iter(time_datasets.keys()))
        else:
            local_time = self.local_time
        return local_time

    def trace_dataset(
        self, dataset: xr.Dataset
    ) -> Tuple[Dict[str, jnp.ndarray], Dict[str, jnp.ndarray], jnp.ndarray]:
        """Trace dataset time indices in order to syncrhonize them an prepare data access
        through unroll operations.

            Args:
                dataset : dataset on which the transformation is applied

            Returns:
                np_data: input data converted in dict of JAX arrays.
                np_index: dict of indices mapping the local time step to the actual indices to access the data.
                xs : range of steps in local time.
        """
        time_dataset = get_time_dataset(dataset)
        time_datasets = split_dataset_from_time_dims(time_dataset)
        local_time = self.get_local_time(time_datasets)

        if len(time_datasets) > 1:
            streams = self.start_dataset_streams(time_datasets)
            stream_merged = self.merge(streams)
            time_index = unroll_stream(stream_merged, pbar=self.pbar)
            # now convert in index for the original dataset.
        else:
            times = time_datasets[local_time][local_time].values
            index = dataset_to_numpy(time_datasets[local_time])
            time_index = {local_time: StreamObservation(times, index)}
        time_dataset_index = get_dataset_index_from_stream_index(time_index)
        dataset_index = get_dataset_index(dataset, time_dataset_index)
        # convert to dict of numpy
        np_data = dataset_to_numpy(dataset)
        np_index = dataset_to_numpy(dataset_index)

        # prepare steps
        xs = jnp.arange(len(time_dataset_index[local_time]))

        return np_data, np_index, xs

    def merge(
        self, streams: Dict[str, Generator[None, None, None]]
    ) -> Generator[None, None, None]:
        """Merge multiple streams in one stream_data.

        Args:
            streams: mapping of streams (generators)

        """
        buffer_dtype_init_values = self.buffer_dtype_init_values or DTYPE_INIT_VALUES
        # initialization
        _output_values: Dict = {}
        stream_obs: Dict = {}
        buffers: Dict = {}
        for time_dim, stream in streams.items():
            # initialization of the generators (first call to next)
            stream_obs[time_dim] = next(stream)

            def get_fill_value(x):
                dtype_ = get_unique_dtype(x)
                return buffer_dtype_init_values[dtype_]

            fill_value = StreamObservation(
                pd.NaT,  # time are initialized to None as there is not nan for datetimes in numpy.
                tree_map(get_fill_value, stream_obs[time_dim].value),
            )

            # init obs
            init_obs = StreamObservation(
                pd.NaT,  # time are initialized to None as there is not nan for datetimes in numpy.
                tree_map(
                    lambda x, fv: onp.full_like(x, fv),
                    stream_obs[time_dim].value,
                    fill_value.value,
                ),
            )

            # fire the buffers
            if time_dim in self.buffer_maxlen:
                buffers[time_dim] = tree_buffer_transform(
                    self.buffer_maxlen[time_dim], fill_value, init_obs
                )
            else:
                # identity transformation
                buffers[time_dim] = lambda x: x

            buffer_output = buffers[time_dim](init_obs)
            _output_values[time_dim] = buffer_output

        generator_state = GeneratorState(_output_values, stream_obs)

        # initialization
        # yield _output_values

        original_streams = streams.copy()
        local_time = self.get_local_time(streams)
        while True:
            # local loop
            streams = original_streams.copy()

            output_values, stream_obs = generator_state

            local_stream = streams.pop(local_time)

            # initialize outputs with initial values
            output = output_values.copy()
            try:
                local_obs = cast(StreamObservation, next(local_stream))
            except StopIteration:
                return

            output[local_time] = local_obs
            local_timestamp = local_obs.time

            if _is_verbose(self.verbose, local_time):
                print(f"[MultiStream] '{local_time}' proceed data : {local_timestamp}")

            for time_dim, stream in streams.items():
                stream_timestamp = stream_obs[time_dim].time

                if time_dim in self.freqs:
                    embed = cast(
                        Callable[[Any], Any],
                        partial(floor_datetime, freq=self.freqs[time_dim]),
                    )
                    _is_before_local = partial(
                        timestamp_is_before_local, embed, local_timestamp
                    )
                else:
                    _is_before_local = partial(
                        timestamp_is_before_local, lambda x: x, local_timestamp
                    )

                if _is_before_local(stream_timestamp):
                    # put last readed observation
                    if _is_verbose(self.verbose, time_dim):
                        print(f"'{time_dim}' proceed data : {stream_timestamp}")
                    buffer_output = buffers[time_dim](stream_obs[time_dim])
                    output[time_dim] = buffer_output

                while _is_before_local(stream_timestamp):
                    try:
                        stream_obs[time_dim] = next(stream)
                        stream_timestamp = stream_obs[time_dim].time
                    except StopIteration:
                        del stream_obs[time_dim]

                        # We will not use anymore this stream_data
                        del original_streams[time_dim]
                        break

                    if _is_before_local(stream_timestamp):
                        buffer_output = buffers[time_dim](stream_obs[time_dim])
                        output[time_dim] = buffer_output
                        if _is_verbose(self.verbose, time_dim):
                            print(f"'{time_dim}' proceed data : {stream_timestamp}")

                if _is_verbose(self.verbose, time_dim):

                    def as_list(vals):

                        vals = onp.array(vals)
                        if vals.ndim == 2:
                            idx = onp.stack([vals[i] for i in range(len(vals))])
                            return pd.DataFrame(output[time_dim].value, index=idx)
                        idx = vals
                        return (output[time_dim].value, idx)

                    print(
                        f"'{time_dim}'  return output: {as_list(output[time_dim].time)}"
                    )

                if self.ffills.get(time_dim, False):
                    # forward fill observation
                    output_values[time_dim] = output[time_dim]

            yield output

    def start_dataset_stream(
        self,
        dataset: xr.Dataset,
        time_dim: Optional[str] = None,
    ):
        """
        Args:
            dataset : xarray Dataset
            time_dim : dimension on witch to iterate
        """
        if time_dim is None:
            time_dim = _get_unique_time_dim(dataset)

        times = dataset.coords[time_dim].values

        # convert to numpy dataset
        np_dataset = dataset_to_numpy(dataset)
        del dataset
        # initialize generator
        i = 0
        time = times[0]
        obs = StreamObservation(time, np_select_item(np_dataset, 0, self.trace))
        if _is_verbose(self.verbose, time_dim):
            print(
                " " * 50 + f"[stream_dataset_generator] Initialization, "
                f"read data '{time_dim}' : index({i})  : '{obs.time}'"
            )
        yield obs

        times = _add_pbar(times, self.pbar, time_dim)
        for i, time in enumerate(times):
            obs = StreamObservation(time, np_select_item(np_dataset, i, self.trace))
            if _is_verbose(self.verbose, time_dim):
                print(
                    " " * 50 + f"[stream_dataset_generator] "
                    f"Read data '{time_dim}' : index({i})  : '{obs.time}'"
                )
            yield obs

    def start_dataset_streams(
        self,
        datasets: Dict[str, xr.Dataset],
    ) -> Dict[str, Generator[None, None, None]]:
        """
        Args:
            datasets: mapping of datasets.

        Returns:
            Dict of generators.
        """
        streams = {
            time_dim: self.start_dataset_stream(tds, time_dim=time_dim)
            for time_dim, tds in datasets.items()
        }
        return streams


def _add_pbar(times, pbar, time_dim):
    if isinstance(pbar, bool):
        if pbar:
            return tqdm(times, desc=time_dim)

    if isinstance(pbar, (list, tuple)):
        if time_dim in pbar:
            return tqdm(times, desc=time_dim)
    return times



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Format nested data structures to numpy/xarray/pandas containers."""
from functools import partial
from typing import Any, TypeVar, Union

import jax.numpy as jnp
import numpy as onp
import pandas as pd
import xarray as xr
from jax.tree_util import tree_flatten, tree_leaves, tree_map, tree_unflatten

DatasetCoordinates = TypeVar(
    "DatasetCoordinates"
)  # xr.core.coordinates.DatasetCoordinates


def _to_dataarray(ref_coords, dataarray, dims):
    coords = {
        ref_dim: ref_coord
        for ref_dim, ref_coord in ref_coords.items()
        if ref_dim in dims
    }
    assert isinstance(
        dataarray, onp.ndarray
    ), "[format_to_dataarray] dataarray should have been converted to numpy array."

    dataarray = xr.DataArray(
        dataarray,
        dims=dims,
        coords=coords,
    )
    return dataarray


def _to_dataframe(ref_coords, dataarray, dims, index_nlevels=1):
    dataarray = _to_dataarray(ref_coords, dataarray, dims)
    if dataarray.ndim == 1:
        return dataarray.to_series()
    elif dataarray.ndim == 2:
        return dataarray.to_pandas()
    else:
        series = dataarray.to_series()
        idx_columns = (onp.arange(index_nlevels, series.index.nlevels)).tolist()
        return series.unstack(idx_columns)


def _to_series(ref_coords, dataarray, dims):
    dataarray = _to_dataarray(ref_coords, dataarray, dims)
    series = dataarray.to_series()
    return series


def _get_dims_from_data(data):
    shape = onp.shape(data)
    dims = tuple(f"dim_{i}" for i in range(len(shape)))
    return dims


def format_dataarray(
    ref_coords: DatasetCoordinates, data: Any, format_dims: Any
) -> Any:
    """Format data following a given data schema

    Args:
        ref_coords : coordinates referential to set data coordinates from format_dims.
        data : Nested data structure with numpy arrays as leaves.
        format_dims : Nested data structure with same treedef than data containing
            dims specification for convertion of leaves to DataArray.

    Returns:
        Nested data structure with same treedef as input data and leaves converted in DataArrays.
    """
    data_flat, treedef = tree_flatten(data)
    format_dims_flat, treedef_format = tree_flatten(format_dims)
    if treedef == treedef_format:
        format_dims_flat = tree_leaves(format_dims_flat)
        vals = tuple(
            _to_dataarray(ref_coords, dataarray, dims)
            for dataarray, dims in zip(data_flat, format_dims_flat)
        )
    else:
        vals = tuple(
            _to_dataarray(ref_coords, dataarray, format_dims) for dataarray in data_flat
        )

    return tree_unflatten(treedef, vals)


def format_dataset(ref_coords: DatasetCoordinates, data: Any, format_dims: Any) -> Any:
    """Format data following a given data schema

    Args:
        ref_coords : coordinates referential to set data coordinates from format_dims.
        data : Nested data structure with numpy arrays as leaves.
        format_dims : Nested data structure with same treedef than data containing
            dims specification for convertion of leaves to DataArray.

    Returns:
        Nested data structure with same treedef as input data and leaves converted in DataArrays.
    """
    output = format_dataarray(ref_coords, data, format_dims)
    try:
        return xr.Dataset(output)
    except TypeError:
        return output


def format_dataframe(
    ref_coords: DatasetCoordinates,
    data: Any,
    format_dims: Any = None,
    index_nlevels: int = 1,
) -> Any:
    """Format data following a given data schema

    Args:
        ref_coords : coordinates referential to set data coordinates from format_dims.
        data : Nested data structure with numpy arrays as leaves.
        format_dims : Nested data structure with same treedef than data containing
            dims specification for convertion of leaves to DataArray.
        index_nlevels : number of levels expected for output index.

    Returns:
        Nested data structure with same treedef as input data and leaves converted in DataArrays.
    """
    ref_coords = {} if ref_coords is None else ref_coords
    data_flat, treedef = tree_flatten(data)
    format_dims_flat, treedef_format = tree_flatten(format_dims)
    if treedef == treedef_format:
        vals = tuple(
            _to_dataframe(ref_coords, dataarray, dims, index_nlevels)
            for dataarray, dims in zip(data_flat, format_dims_flat)
        )
    else:
        if not format_dims:
            format_dims = _get_dims_from_data(data_flat[0])
        vals = tuple(
            _to_dataframe(ref_coords, dataarray, format_dims, index_nlevels)
            for dataarray in data_flat
        )

    return tree_unflatten(treedef, vals)


def format_series(ref_coords: DatasetCoordinates, data: Any, format_dims: Any) -> Any:
    """Format data following a given data schema

    Args:
        ref_coords : coordinates referential to set data coordinates from format_dims.
        data : Nested data structure with numpy arrays as leaves.
        format_dims : Nested data structure with same treedef than data containing
            dims specification for convertion of leaves to DataArray.

    Returns:
        Nested data structure with same treedef as input data and leaves converted in DataArrays.
    """
    data_flat, treedef = tree_flatten(data)
    format_dims_flat, treedef_format = tree_flatten(format_dims)
    if treedef == treedef_format:
        vals = tuple(
            _to_series(ref_coords, dataarray, dims)
            for dataarray, dims in zip(data_flat, format_dims_flat)
        )
    else:
        vals = tuple(
            _to_series(ref_coords, dataarray, format_dims) for dataarray in data_flat
        )
    return tree_unflatten(treedef, vals)


def auto_format_with_shape(
    dataset: xr.Dataset, data: jnp.ndarray, output_format: str = "dataarray"
) -> Union[jnp.ndarray, pd.DataFrame, xr.DataArray]:
    """Format data structure in pandas / xarray DataFrame / DataArray using coordinates
    of variables with similar shapes in a reference Dataset.

    :param dataset: template dataset
    :param data: data to format
    :param output_format: format type, can be in ("dataarray", "pandas")
    :return: formated data
    """

    def get_dims():
        for k, v in dataset.items():
            for i in range(len(v.shape) + 1):
                shape = tuple(v.shape[:i])
                if data.shape == shape:
                    return v.dims[:i]
        print(f"The dataset does not have variable with shape {data.shape}.")

    dims = get_dims()
    if dims:
        if output_format == "dataarray":
            data = format_dataarray(dataset.coords, onp.array(data), format_dims=dims)
        elif output_format == "dataframe":
            data = format_dataframe(dataset.coords, onp.array(data), format_dims=dims)
        else:
            raise ValueError(f"Unknown format {output_format}.")
    return data


def tree_auto_format_with_shape(
    dataset: xr.Dataset, pytree: Any, output_format: str = "dataarray"
) -> Any:
    """Format data structure in pandas / xarray DataFrame / DataArray using coordinates
    of variables with similar shapes in a reference Dataset.

    :param dataset: template dataset
    :param data: data to format
    :param output_format: format type, can be in ("dataarray", "pandas")
    :return: formated data
    """
    return tree_map(
        partial(auto_format_with_shape, dataset, output_format=output_format), pytree
    )



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import pytest

from wax.modules.apply_mask import ApplyMask


def test_apply_mask_axis_0():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3))

    mask = jnp.full(x.shape[0], True)
    mask = mask.at[0].set(False)

    # check that an error is raised if axis is not specified
    fun = hk.transform(lambda mask, x: ApplyMask()(mask, x))
    with pytest.raises(ValueError):
        fun.init(next(rng), mask, x)

    # check that specifying the wrong axis raises
    fun = hk.transform(lambda mask, x: ApplyMask(axis=1)(mask, x))
    with pytest.raises(ValueError):
        fun.init(next(rng), mask, x)

    # now specify axis
    fun = hk.transform(lambda mask, x: ApplyMask(axis=0)(mask, x))
    params = fun.init(next(rng), mask, x)
    x_mask = fun.apply(params, next(rng), mask, x)

    assert jnp.allclose(x_mask[0, :], 0.0)


def test_apply_mask_axis_1():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3))

    mask = jnp.full(x.shape[1], True)
    mask = mask.at[0].set(False)

    # check that an error is raised if axis is not specified
    fun = hk.transform(lambda mask, x: ApplyMask()(mask, x))
    with pytest.raises(ValueError):
        fun.init(next(rng), mask, x)

    # check that specifying the wrong axis raises
    fun = hk.transform(lambda mask, x: ApplyMask(axis=0)(mask, x))
    with pytest.raises(ValueError):
        fun.init(next(rng), mask, x)

    # now specify axis
    fun = hk.transform(lambda mask, x: ApplyMask(axis=1)(mask, x))
    params = fun.init(next(rng), mask, x)
    x_mask = fun.apply(params, next(rng), mask, x)

    assert jnp.allclose(x_mask[:, 0], 0.0)


def test_apply_mask_axis_2_raises():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3, 1))

    mask = jnp.full(x.shape[1], True)
    mask = mask.at[0].set(False)

    # check that an error is raised if axis is not specified
    fun = hk.transform(lambda mask, x: ApplyMask(axis=2)(mask, x))
    with pytest.raises(ValueError) as err:
        fun.init(next(rng), mask, x)
    assert (
        str(err.value)
        == "ApplyMask is not implemented for axis different from None, 0, 1."
    )



# Copyright 2022 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable

import haiku as hk
import jax
from haiku.experimental import lift_with_state
from jax.tree_util import tree_map


def vmap_lift_with_state(fun: Callable, split_rng=False, init_rng=True):
    if split_rng and not init_rng:
        raise ValueError("split_rng=True requires init_rng=True")

    def apply_fn(*args, **kwargs):

        tfun = hk.transform_with_state(fun)

        if not split_rng:
            rng = hk.next_rng_key() if (hk.running_init() and init_rng) else None

            @jax.vmap
            def init_fn(*args, **kwargs):
                return tfun.init(rng, *args, **kwargs)

            params_and_state_fn, updater = hk.experimental.lift_with_state(
                init_fn, name="vmap_lift"
            )
            params, state = params_and_state_fn(*args, **kwargs)

            @jax.vmap
            def apply_fn(params, state, *args, **kwargs):
                return tfun.apply(params, state, None, *args, **kwargs)

            out, state = apply_fn(params, state, *args, **kwargs)
        else:
            n_batches = len(jax.tree_util.tree_leaves((args, kwargs))[0])
            if init_rng and hk.running_init():
                rng = hk.next_rng_key()
                rng = jax.random.split(rng, num=n_batches)
            else:
                rng = None

            @jax.vmap
            def init_fn(rng, *args, **kwargs):
                return tfun.init(rng, *args, **kwargs)

            params_and_state_fn, updater = hk.experimental.lift_with_state(
                init_fn, name="vmap_lift"
            )
            params, state = params_and_state_fn(rng, *args, **kwargs)

            @jax.vmap
            def apply_fn(params, state, rng, *args, **kwargs):
                return tfun.apply(params, state, rng, *args, **kwargs)

            rng = hk.next_rng_key()
            rng = jax.random.split(rng, num=n_batches)
            out, state = apply_fn(params, state, rng, *args, **kwargs)

        updater.update(state)
        return out

    return apply_fn


def unroll_lift_with_state(
    fn: Callable, skip_first=False, split_rng=False, init_rng=True
):
    if split_rng and not init_rng:
        raise ValueError("split_rng=True requires init_rng=True")

    def apply_fn(*args, **kwargs):
        tfn = hk.transform_with_state(fn)
        params_and_state_fn, updater = lift_with_state(tfn.init, name="unroll_lift")

        def init(xs):
            args_0, kwargs_0 = tree_map(lambda x: x[0], xs)
            rng = hk.next_rng_key() if (hk.running_init() and init_rng) else None
            params, state = params_and_state_fn(rng, *args_0, **kwargs_0)
            return params, state

        def scan_f(state, inputs):
            args_step, kwargs_step = inputs
            rng = hk.maybe_next_rng_key() if split_rng else None
            outputs, state = tfn.apply(params, state, rng, *args_step, **kwargs_step)
            return state, outputs

        xs = (args, kwargs)
        params, state = init(xs)
        if skip_first:
            xs = tree_map(lambda x: x[1:], xs)
        final_state, output_sequence = hk.scan(scan_f, init=state, xs=xs)
        updater.update(final_state)

        return output_sequence

    return apply_fn



# Copyright 2022 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import optax
from jax.config import config

from wax.modules.func_optimizer import FuncOptimizer
from wax.modules.optax_optimizer import OptaxOptimizer
from wax.unroll import unroll


def generate_data():
    T = 1000
    rng = jax.random.PRNGKey(42)
    x = jax.random.normal(rng, (T, 3))
    w = jnp.ones((3,)).at[2].set(0)
    y = x @ w + 0.1 * jax.random.normal(rng, (T,))
    return x, y


def test_func_optimizer():
    config.update("jax_enable_x64", False)

    x, y = generate_data()

    def loss(y, yp):
        return jnp.square(y - yp).mean()

    def learn(x, y):
        def cost(x, y):
            yp = hk.Linear(1, with_bias=False)(x)
            return loss(y, yp.reshape(y.shape)), yp

        (l, yp), params = FuncOptimizer(
            cost, OptaxOptimizer(optax.sgd(1.0e-2)), has_aux=True
        )(x, y)

        return (l, yp), params

    rng = jax.random.PRNGKey(42)
    res = unroll(learn, rng=rng)(x, y)
    (l_, yp), w_history = res

    # check that averaged loss is less than initial loss.
    assert l_.mean() < l_[0] / 37
    # pd.DataFrame(w_history["linear"]["w"].squeeze()).plot()
    # plt.show();
    # pd.Series(l).expanding().mean().plot(title="loss")



# Copyright 2022 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import haiku as hk
import jax
import jax.numpy as jnp
import pytest

from wax.stateful import unroll_lift_with_state, vmap_lift_with_state


class MyModule(hk.Module):
    def __init__(self, steps, name=None):
        super().__init__(name=name)
        self.steps = steps

    def __call__(self, x):
        assert x.ndim == 0
        steps = hk.get_parameter(
            "steps", (1,), init=lambda *_: self.steps * jnp.ones(1)
        )
        state = hk.get_state("state", (1,), init=lambda *_: jnp.zeros(1) + x)
        state = state + steps
        hk.set_state("state", state)
        return state + x


@pytest.mark.parametrize(
    "init_rng, split_rng", [[False, False], [True, False], [True, True]]
)
def test_vmap_lift_wtih_state(init_rng, split_rng):

    x = jnp.arange(3).astype(jnp.float32)

    def run_vmap():
        def outer_fun(x):
            def fun(x):
                return MyModule(steps=2)(x)

            return vmap_lift_with_state(fun, split_rng=split_rng, init_rng=init_rng)(x)

        init, apply = hk.transform_with_state(outer_fun)
        if init_rng or split_rng:
            params, state = init(jax.random.PRNGKey(0), x)
        else:
            params, state = init(None, x)

        if split_rng:
            rng = jax.random.PRNGKey(0)
            out, state = apply(params, state, rng, x)
            out, state = apply(params, state, rng, x)
        else:
            out, state = apply(params, state, None, x)
            out, state = apply(params, state, None, x)
        return out

    def run_static():
        steps = 2
        state = jnp.zeros_like(x) + x

        # apply 1
        state = state + steps
        out = state + x
        # apply 2
        state = state + steps
        out = state + x
        return out.reshape(-1, 1)

    assert jnp.allclose(run_vmap(), run_static())


@pytest.mark.parametrize(
    "init_rng, split_rng", [[False, False], [True, False], [True, True]]
)
def test_unroll_lift_wtih_state(init_rng, split_rng):
    x = jnp.zeros(3).astype(jnp.float32)

    def run_unroll():
        def outer_fun(x):
            def fun(x):
                return MyModule(steps=1)(x)

            return unroll_lift_with_state(fun, init_rng=init_rng, split_rng=split_rng)(
                x
            )

        init, apply = hk.transform_with_state(outer_fun)
        if init_rng or split_rng:
            params, state = init(jax.random.PRNGKey(0), x)
        else:
            params, state = init(None, x)
        if split_rng:
            rng = jax.random.PRNGKey(0)
            out, state = apply(params, state, rng, x)
            out, state = apply(params, state, rng, x)
        else:
            out, state = apply(params, state, None, x)
            out, state = apply(params, state, None, x)
        return out

    def run_static():
        steps = 1
        state = jnp.zeros(1) + x[0]
        out = []
        for i in range(len(x)):
            state = state + steps
            out.append(state + x[i])
        out = []
        for i in range(len(x)):
            state = state + steps
            out.append(state + x[i])
        out = jnp.stack(out)
        return out

    assert jnp.allclose(run_unroll(), run_static())



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import haiku as hk
import jax
import jax.numpy as jnp
import pytest
from haiku._src.data_structures import FlatMapping
from jax.config import config
from jax.numpy import array as DeviceArray
from jax.numpy import int32, uint32

from wax.modules.counter import Counter
from wax.modules.gym_feedback import GymFeedback, GymOutput
from wax.testing import assert_tree_all_close
from wax.transform import (
    BatchState,
    transform_batch_with_state,
    transform_batch_with_state_static,
)
from wax.unroll import data_unroll, gym_static_unroll, unroll

GymState = GymFeedback.GymState


def raw_observations():
    for i in range(10):
        yield i, {"some_information_about_data"}


class Agent(hk.Module):
    def __call__(self, obs):
        action = obs + 1
        return {"structured_action": action}, {"agent_info": Counter()()}


class Env(hk.Module):
    def __call__(self, action, raw_obs):
        obs = raw_obs + action["structured_action"]
        rw = action["structured_action"] * obs
        return rw, obs, {"env_info": Counter()()}


def test_gym_module_gym_static_unroll():
    config.update("jax_enable_x64", False)

    @hk.transform_with_state
    def gym_fun(raw_obs):
        return GymFeedback(Agent(), Env(), return_obs=True, return_action=True)(raw_obs)

    raw_generator = iter(raw_observations())
    seq = hk.PRNGSequence(42)
    (gym_output, gym_info), state = gym_static_unroll(
        gym_fun, None, None, seq, True, raw_generator
    )

    ref_gym_output = GymOutput(return_obs=True, return_action=True).format(
        reward=DeviceArray([2, 15, 54, 140, 300, 567, 980, 1584, 2430], dtype=int32),
        obs=DeviceArray([2, 5, 9, 14, 20, 27, 35, 44, 54], dtype=int32),
        action={
            "structured_action": DeviceArray(
                [3, 6, 10, 15, 21, 28, 36, 45, 55], dtype=int32
            )
        },
    )

    ref_state = FlatMapping(
        {
            "gym_feedback": FlatMapping(
                {"action": GymState(action={"structured_action": 55})}
            ),
            "agent/counter": FlatMapping({"count": DeviceArray(9, dtype=uint32)}),
            "env/counter": FlatMapping({"count": DeviceArray(9, dtype=uint32)}),
        }
    )
    assert_tree_all_close(state, ref_state)
    assert_tree_all_close(gym_output, ref_gym_output)


def test_gym_module_dynamic_unroll():
    config.update("jax_enable_x64", False)

    @hk.transform_with_state
    def gym_fun(raw_obs):
        return GymFeedback(Agent(), Env())(raw_obs)

    xs = data_unroll(iter(raw_observations()))
    rng = next(hk.PRNGSequence(42))
    (gym_output, gym_info), final_state = unroll(
        gym_fun, return_final_state=True, rng=rng, skip_first=True
    )(xs)

    # reference outputs

    ref_final_state = FlatMapping(
        {
            "agent/counter": FlatMapping({"count": DeviceArray(9, dtype=uint32)}),
            "env/counter": FlatMapping({"count": DeviceArray(9, dtype=uint32)}),
            "gym_feedback": FlatMapping(
                {
                    "action": GymState(
                        action={"structured_action": DeviceArray(55, dtype=int32)}
                    ),
                }
            ),
        }
    )

    ref_output_sequence = GymOutput().format(
        reward=DeviceArray([2, 15, 54, 140, 300, 567, 980, 1584, 2430], dtype=int32),
    )

    assert_tree_all_close(final_state, ref_final_state)
    assert_tree_all_close(gym_output, ref_output_sequence)


@pytest.mark.parametrize("static", [True, False])
def test_gym_module_transform_batch_with_state(static):
    config.update("jax_enable_x64", False)

    @hk.transform_with_state
    def gym_fun(raw_obs):
        return GymFeedback(Agent(), Env())(raw_obs)

    xs = data_unroll(iter(raw_observations()))

    rng = next(hk.PRNGSequence(42))
    if static:
        xs, rng = jax.device_put((xs, rng))
        batch_fun = transform_batch_with_state_static(gym_fun, skip_first=True)
    else:
        batch_fun = transform_batch_with_state(gym_fun, skip_first=True)
    batch_params, batch_state = batch_fun.init(rng, xs)
    print("first action: ", batch_state.fun_state)

    (output_sequence, info), batch_state = batch_fun.apply(
        batch_params, batch_state, rng, xs
    )

    assert batch_state._fields == ("fun_params", "fun_state", "rng_key")
    assert output_sequence._fields == ("reward",)

    assert (
        output_sequence.reward
        == jnp.array([2, 15, 54, 140, 300, 567, 980, 1584, 2430], dtype=jnp.int32)
    ).all()

    ref_output_sequence = GymOutput().format(
        reward=DeviceArray([2, 15, 54, 140, 300, 567, 980, 1584, 2430], dtype=int32),
    )
    ref_batch_state = BatchState(
        fun_params=FlatMapping({}),
        fun_state=FlatMapping(
            {
                "gym_feedback": FlatMapping(
                    {
                        "action": GymState(
                            action={"structured_action": DeviceArray(55, dtype=int32)}
                        ),
                    }
                ),
                "agent/counter": FlatMapping({"count": DeviceArray([9], dtype=uint32)}),
                "env/counter": FlatMapping({"count": DeviceArray([9], dtype=uint32)}),
            }
        ),
        rng_key=DeviceArray([255383827, 267815257], dtype=uint32),
    )

    assert_tree_all_close(batch_state, ref_batch_state)
    assert_tree_all_close(output_sequence, ref_output_sequence)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fill nan, posinf and neginf values."""
from typing import Any, Optional

import haiku as hk
import jax.numpy as jnp
from jax import tree_map


class FillNanInf(hk.Module):
    """Fill nan, posinf and neginf values."""

    def __init__(self, fill_value: Any = 0.0, name: Optional[str] = None):
        """Initialize module.

        Args:
            fill_value : value used to replace nan, posinf or neginf encountered values.
            name : name of the module
        """
        super().__init__(name=name)
        self.fill_value = fill_value

    def __call__(self, input):
        def fill_nan(x):
            # mask values
            return jnp.nan_to_num(
                x, nan=self.fill_value, posinf=self.fill_value, neginf=self.fill_value
            )

        return tree_map(fill_nan, input)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax.numpy as jnp

from wax.compile import jit_init_apply
from wax.modules.update_on_event import UpdateOnEvent

TEST_KEY = jnp.array([255383827, 267815257], dtype=jnp.uint32)


class MyModule(hk.Module):
    def __call__(self, x):
        prev_state = hk.get_state(
            "state",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, fill_value=0, dtype=dtype),
        )

        state = prev_state
        state = state + x
        output = state * 2

        hk.set_state("state", state)

        return output


def test_no_wrapping():
    @jit_init_apply
    @hk.transform_with_state
    def fun(x):
        return MyModule()(x)

    x = jnp.array(
        [
            -1.6652094,
            0.23443973,
            -0.24498996,
            0.4470721,
            0.25630304,
            0.28259817,
            -0.11017023,
            0.5699526,
            0.33960477,
            -1.7104211,
        ],
        dtype=jnp.float32,
    )
    params, state = fun.init(TEST_KEY, x)

    outputs = []
    for i in range(10):
        output, state = fun.apply(params, state, TEST_KEY, x)
        outputs.append(output)
    outputs = jnp.stack(outputs)
    ref_outputs = jnp.array(
        [
            -3.3304188,
            -6.6608377,
            -9.991257,
            -13.321675,
            -16.652094,
            -19.982513,
            -23.312933,
            -26.643353,
            -29.973772,
            -33.30419,
        ],
        dtype=jnp.float32,
    )
    assert jnp.allclose(ref_outputs, outputs[:, 0])


def test_deleguate_all_event_true():
    @jit_init_apply
    @hk.transform_with_state
    def fun(event, x):
        return UpdateOnEvent(MyModule())(event, x)

    x = jnp.array(
        [
            -1.6652094,
            0.23443973,
            -0.24498996,
            0.4470721,
            0.25630304,
            0.28259817,
            -0.11017023,
            0.5699526,
            0.33960477,
            -1.7104211,
        ],
        dtype=jnp.float32,
    )

    event = True
    params, state = fun.init(TEST_KEY, event, x)
    outputs = []
    for i in range(10):
        output, state = fun.apply(params, state, TEST_KEY, event, x)
        outputs.append(output)
    outputs = jnp.stack(outputs)
    ref_outputs = jnp.array(
        [
            -3.3304188,
            -6.6608377,
            -9.991257,
            -13.321675,
            -16.652094,
            -19.982513,
            -23.312933,
            -26.643353,
            -29.973772,
            -33.30419,
        ],
        dtype=jnp.float32,
    )
    assert jnp.allclose(ref_outputs, outputs[:, 0])


def test_deleguate_some_event_true():
    @jit_init_apply
    @hk.transform_with_state
    def fun(event, x):
        return UpdateOnEvent(MyModule(), initial_output_value=-999)(event, x)

    x = jnp.array(
        [
            -1.6652094,
            0.23443973,
            -0.24498996,
            0.4470721,
            0.25630304,
            0.28259817,
            -0.11017023,
            0.5699526,
            0.33960477,
            -1.7104211,
        ],
        dtype=jnp.float32,
    )

    event = False
    params, state = fun.init(TEST_KEY, event, x)

    outputs = []
    for i in range(10):
        if i in [3, 7]:
            event = True
        else:
            event = False
        output, state = fun.apply(params, state, TEST_KEY, event, x)
        outputs.append(output)
    outputs = jnp.stack(outputs)
    ref_outputs = jnp.array(
        [
            -999.0,
            -999.0,
            -999.0,
            -3.3304188,
            -3.3304188,
            -3.3304188,
            -3.3304188,
            -6.6608377,
            -6.6608377,
            -6.6608377,
        ],
        dtype=jnp.float32,
    )
    assert jnp.allclose(ref_outputs, outputs[:, 0])



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial

import haiku as hk
import numpy as np
import numpy as onp
import pandas as pd
import xarray as xr

from wax.gym.callbacks.progressbar import ProgressBar
from wax.gym.callbacks.record import Record
from wax.gym.callbacks.stop import Stop
from wax.gym.haiku_agent import HaikuAgent
from wax.gym.haiku_env import HaikuEnv


def gym_unroll(env, agent):
    obs = env.reset()
    agent.reset()
    while True:
        action = agent(obs)
        obs, rw, done, info = env.step(action)
        print(action, obs, rw, done, info)
        if done:
            break
    return action, obs, rw, done, info


def raw_observations():
    for i in range(10):
        yield i, {"some_information_about_data"}


def raw_observations_array():
    for i in range(10):
        yield onp.full((1, 1), i), {"some_information_about_data"}


def raw_observations_dataframe():
    for i in range(10):
        yield pd.DataFrame(onp.full((1, 1), i)), {"some_information_about_data"}


def raw_observations_dataarray():
    for i in range(10):
        yield xr.DataArray(onp.full((1, 1), i)), {"some_information_about_data"}


class Agent(hk.Module):
    def __call__(self, obs):
        action = obs + 1
        return {"structured_action": action}


class Env(hk.Module):
    def __call__(self, action, raw_obs):
        obs = raw_obs + action["structured_action"]
        rw = action["structured_action"] * obs
        return rw, obs


def test_haiku_agent_and_env():
    data_generator = iter(raw_observations())

    @partial(HaikuEnv, data_generator)
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @HaikuAgent
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    action, obs, rw, done, info = gym_unroll(env, agent)
    assert action["structured_action"] == 55


def test_haiku_agent_and_env2():
    @partial(HaikuEnv, iter(raw_observations()), name="Env")
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @partial(HaikuAgent, name="Agent")
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    action, obs, rw, done, info = gym_unroll(env, agent)

    assert action["structured_action"] == 55


def test_unroll():
    @partial(HaikuEnv, iter(raw_observations()), name="Env")
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @partial(HaikuAgent, name="Agent")
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    with Record() as rec:
        state = agent.unroll(env)

    assert state.action["structured_action"] == 55
    outputs = rec.get_outputs()
    assert isinstance(outputs.reward, onp.ndarray)
    assert (
        outputs.reward == onp.array([2, 15, 54, 140, 300, 567, 980, 1584, 2430])
    ).all()


def test_unroll_array():
    @partial(HaikuEnv, iter(raw_observations_array()), name="Env")
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @partial(HaikuAgent, name="Agent")
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    with Record() as rec, ProgressBar():

        state = agent.unroll(env)

    assert isinstance(state.action["structured_action"], np.ndarray)
    assert state.action["structured_action"] == 55

    outputs = rec.get_outputs()
    assert isinstance(outputs.reward, np.ndarray)
    assert outputs.reward.tolist() == (
        [[[2]], [[15]], [[54]], [[140]], [[300]], [[567]], [[980]], [[1584]], [[2430]]]
    )


def test_unroll_dataframe():
    @partial(HaikuEnv, iter(raw_observations_dataframe()), name="Env")
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @partial(HaikuAgent, name="Agent")
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    with Record() as rec, ProgressBar():

        state = agent.unroll(env)

    assert isinstance(state.action["structured_action"], pd.DataFrame)
    assert state.action["structured_action"].values == 55

    outputs = rec.get_outputs()
    assert isinstance(outputs.reward, pd.DataFrame)
    assert outputs.reward.values.tolist() == (
        [[2], [15], [54], [140], [300], [567], [980], [1584], [2430]]
    )


def test_unroll_dataframe_stop_callback():
    @partial(HaikuEnv, iter(raw_observations_dataframe()), name="Env")
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @partial(HaikuAgent, name="Agent")
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    with Record() as rec, ProgressBar(), Stop(3):

        state = agent.unroll(env)

    assert isinstance(state.action["structured_action"], pd.DataFrame)
    assert state.action["structured_action"].values == 6

    outputs = rec.get_outputs()
    assert isinstance(outputs.reward, pd.DataFrame)
    assert outputs.reward.values.tolist() == ([[2], [15], [54]])


def test_unroll_dataarray_stop_callback():
    @partial(HaikuEnv, iter(raw_observations_dataarray()), name="Env")
    @hk.transform_with_state
    def env(action, raw_obs):
        return Env()(action, raw_obs)

    @partial(HaikuAgent, name="Agent")
    @hk.transform_with_state
    def agent(obs):
        return Agent()(obs)

    with Record() as rec, ProgressBar(), Stop(3):

        state = agent.unroll(env)

    assert isinstance(state.action["structured_action"], xr.DataArray)
    assert state.action["structured_action"].values == 6

    outputs = rec.get_outputs()
    assert isinstance(outputs.reward, xr.DataArray)
    assert outputs.reward.values.tolist() == ([[2], [15], [54]])



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Encoding schemes to encode/decode numpy data types non supported by JAX, e.g.
`datetime64` and `string_` dtypes."""
import datetime
from typing import Any, Callable, Dict, NamedTuple, TypeVar, Union

import numpy as onp
import pandas as pd
from jax import numpy as jnp
from pandas.tseries.offsets import DateOffset
from sklearn.preprocessing import LabelEncoder

DTypeLike = TypeVar("DTypeLike")


def encode_int64(seed: int) -> onp.ndarray:
    """Encode an int64 into in a 2-dimensional int32 ndarray.

    Args:
      seed: a 64- or 32-bit integer.

    Returns:
      an array of shape (2,) and dtype uint32.

    References
        See
        - jax implementation of PRNGKey.
        https://jax.readthedocs.io/en/latest/_autosummary/jax.random.PRNGKey.html  # noqa
        - https://codereview.stackexchange.com/questions/80386/packing-and-unpacking-two-32-bit-integers-into-an-unsigned-64-bit-integer  # noqa
    Note:
        0xFFFFFFFF = 2**32 -1
    """
    if onp.shape(seed):
        raise TypeError("seed must be a scalar.")
    if isinstance(seed, onp.ndarray):
        seed = seed.item()
    if not isinstance(seed, (int, onp.int32, onp.uint32, onp.int64)):
        raise TypeError(f"seed must be an int, got {type(seed)}")

    def _convert(k):
        return onp.reshape(k.astype(onp.uint32), [1])

    if isinstance(seed, (int, onp.ndarray, onp.int32, onp.uint32)):
        # Special handling of raw integer values, which may have be 64bit even
        # when jax_enable_x64=False and we don't want to drop the top 32 bits
        high = _convert(onp.bitwise_and(onp.right_shift(seed, 32), 0xFFFFFFFF))
    else:
        high = _convert(onp.right_shift(seed, onp.full_like(seed, 32)))
    low = _convert(onp.bitwise_and(seed, 0xFFFFFFFF))
    return onp.concatenate([high, low], 0)


def decode_int64(code: Union[onp.ndarray, jnp.ndarray]) -> onp.ndarray:
    """See https://codereview.stackexchange.com/questions/80386/packing-and-unpacking-two-32-bit-integers-into-an-unsigned-64-bit-integer  # noqa"""
    # assert isinstance(code, np.ndarray)
    high, low = onp.array(code)

    def _convert(k):
        return onp.reshape(k.astype(onp.int64), [1])

    high = _convert(onp.left_shift(high, 32))
    low = _convert(low)
    return low + high


def encode_timestamp(date):
    if not isinstance(date, pd.Timestamp):
        raise TypeError(f"seed must be a pandas Timestamp, got {type(date)}")
    return encode_int64(date.value)


def decode_timestamp(code):
    return pd.to_datetime(decode_int64(code)[0])


def encode_datetime(date):
    if not isinstance(date, datetime.datetime):
        raise TypeError(f"seed must be a datetime, got {type(date)}")
    return encode_datetime64(onp.datetime64(date))


def decode_datetime(code):
    return pd.to_datetime(decode_int64(code)[0]).to_pydatetime()


def encode_datetime64(date):
    if date.dtype not in [
        onp.dtype("datetime64[ns]"),
        onp.dtype("<M8[ns]"),
        onp.dtype("<M8[us]"),
    ]:
        raise TypeError(f"seed must be a np.datetime64, got {type(date)}")
    return encode_int64(date.astype("<M8[ns]").astype(onp.int64))


def decode_datetime64(code):
    return decode_int64(code)[0].astype("<M8[ns]")


class Encoder(NamedTuple):
    encode: Callable
    decode: Callable


class DecodeDataset(NamedTuple):
    iter_coords: Dict
    embed_map: Dict
    event_map: Dict
    embedding_coords: Dict
    event_coords: Dict
    other_coords: Dict
    encoders: Dict
    constant_data: Dict
    iter_data: Dict
    embed_data: Dict
    event_data: Dict


def floor_datetime(time, freq: Union[str, DateOffset] = "d"):
    r"""Perform floor operation on the data to the specified freq.
    Args:
        time : time(s) to floor.
        freq : The frequency level to floor the index to. Must be a fixed frequency
            like ‘S’ (second) not ‘ME’ (month end).
            See frequency aliases for a list of possible freq values.

    See:
    https://pandas.pydata.org/docs/reference/api/pandas.Series.dt.floor.html

    As discussed in:
    https://stackoverflow.com/questions/5476065/how-to-truncate-the-time-on-a-datetime-object-in-python  # noqa
    the implementation with pd.Series.dt.florr seems to be the most performant.

        times = pd.Series(pd.date_range(start='1/1/2018 04:00:00', end='1/1/2018 22:00:00', freq='s'))
        %timeit floor_datetime(times)
        791 µs ± 4.19 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

    https://docs.oracle.com/cd/B19306_01/server.102/b14200/functions201.htm
    """
    assert freq == "d", f"other formats than '{freq}' not yet supported."
    # v3 (only in numpy)
    time_int64 = time.astype("<M8[ns]").astype(onp.int64)
    day = onp.array(24 * 60 * 60 * 1.0e9, onp.int64)
    floor_time_int64 = onp.array((time_int64 // day) * day, onp.int64)
    floor_time = floor_time_int64.astype("<M8[ns]")
    return floor_time

    # v1
    # assert freq == "d", f"other formats than '{freq}' not yet supported"
    # return pd.to_datetime(time).round(freq).to_numpy().astype("<M8[ns]")

    # v2
    # return pd.Series(time).dt.floor(freq).values.reshape(time.shape).astype("<M8[ns]")


def string_encoder(values: Any) -> Encoder:

    ravel = False
    original_shape = None

    if values.ndim > 1:
        ravel = True
        original_shape = values.shape
        values = values.ravel()
    encoder = LabelEncoder().fit(values)

    def encode(value):
        if ravel:
            value = value.ravel()
        return encoder.transform(value)

    def decode(code):
        value = encoder.inverse_transform(code)
        if ravel:
            value = value.reshape(original_shape)
        return value

    return Encoder(encode, decode)


def datetime64_encoder(values: Any) -> Encoder:
    def encode(value):
        return onp.stack(list(map(encode_datetime64, value)))

    def decode(code):
        return onp.stack(
            list(map(decode_datetime64, [code[i] for i in range(len(code))]))
        )

    return Encoder(encode, decode)


def less(t1, t2):
    h1, l1 = t1
    h2, l2 = t2
    if h1 < h2:
        return True
    elif h1 > h2:
        return False
    else:
        if l1 < l2:
            return True
        else:
            return False


def floor_jax_datekey(datekey):
    """Floor a date represented as a datekey.

    TODO: find an efficient implementation which do not need
    to pass by int64 conversions in order to use it smoothly
    in Jax 32 bits worflows.

    """

    date = decode_int64(onp.array(datekey))
    day = onp.array(24 * 60 * 60 * 1.0e9, onp.int64)
    floor_date = onp.ndarray.item((date // day) * day)
    datekey = encode_int64(floor_date)
    datekey = jnp.array(datekey)
    return datekey


def encode_dataset(encoders, dataset):
    output = {}
    for dim, var in dataset.items():
        if dim in encoders:
            # encode values
            values = encoders[dim].encode(var)
        else:
            values = var
        output[dim] = values
    return output


def decode_dataset(encoders, dataset):
    output = {}
    for dim, var in dataset.items():
        if dim in encoders:
            # encode values
            values = encoders[dim].decode(var)
        else:
            values = var
        output[dim] = values
    return output



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as onp
import xarray as xr
from jax.config import config

from wax.accessors import register_wax_accessors
from wax.datasets.generate_temperature_data import generate_temperature_data
from wax.modules.has_changed import HasChanged
from wax.modules.ohlc import OHLC


def test_ohlc_temperature():
    register_wax_accessors()
    config.update("jax_enable_x64", True)

    # da = xr.tutorial.open_dataset("air_temperature").sel(time="2013-01")
    da = generate_temperature_data()
    da["date"] = da.time.dt.floor("d").astype(onp.datetime64)
    assert len(set(da.date.values.tolist())) == 6
    assert len(set(da.time.values.tolist())) == 124

    def bin_temperature(da):
        date_has_changed = HasChanged()(da["date"])
        return OHLC()(da["air"], reset_on=date_has_changed)

    output, state = da.wax.stream().apply(bin_temperature, format_dims=da.air.dims)
    output = xr.Dataset(output._asdict())
    df = output.isel(lat=0, lon=0).drop_vars(["lat", "lon"]).to_array().to_pandas().T
    assert (df["HIGH"] >= df["LOW"]).all()
    assert (df["HIGH"] >= df["CLOSE"]).all()
    assert (df["HIGH"] >= df["OPEN"]).all()
    assert (df["CLOSE"] >= df["LOW"]).all()
    assert (df["OPEN"] >= df["LOW"]).all()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Online supervised learner."""
from typing import Any, Callable, NamedTuple, Optional, Tuple, Union

import haiku as hk
import jax.numpy as jnp
from optax import GradientTransformation

from wax.modules.func_optimizer import FuncOptimizer
from wax.modules.optax_optimizer import OptaxOptimizer


class ParamsState(NamedTuple):
    params: Any
    state: Any


class OnlineSupervisedLearnerInfo(NamedTuple):
    loss: float
    params: Any


class OnlineSupervisedLearner(hk.Module):
    """Online supervised learner."""

    def __init__(
        self,
        model: Callable,
        opt: Union[OptaxOptimizer, GradientTransformation],
        loss: Callable,
        grads_fill_nan_inf=True,
        name: Optional[str] = None,
    ):
        """Initialize module.

        Args:
            model : model to optimize.
            opt : optimizer.
            loss : loss function.
            name : name of the module
            grads_fill_nan_inf: if true, fill nan and
                +/- infinite values in gradients with zeros.
        """
        super().__init__(name=name)
        self.model = model
        self.opt = (
            OptaxOptimizer(opt) if isinstance(opt, GradientTransformation) else opt
        )
        self.loss = loss
        self.grads_fill_nan_inf = grads_fill_nan_inf

    def __call__(
        self, x: jnp.ndarray, y: jnp.ndarray
    ) -> Tuple[jnp.ndarray, OnlineSupervisedLearnerInfo]:
        """Update learner.

        Args:
            x : features
            y: target
        """

        def _loss(x, y):
            y_pred = self.model(x)
            return self.loss(y, y_pred), y_pred

        (l, y_pred), params = FuncOptimizer(
            _loss, self.opt, has_aux=True, grads_fill_nan_inf=self.grads_fill_nan_inf
        )(x, y)
        info = OnlineSupervisedLearnerInfo(loss=l, params=params)
        return y_pred, info



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Some utils functions used in WAX-ML."""

from jax.tree_util import tree_flatten


def dict_map(fun, col):
    return {key: fun(val) for key, val in col.items()}


def get_unique_dtype(current_values_):
    # check of unique dtype
    # TODO remove onece multi-dtype is supported
    current_values_flat_, _ = tree_flatten(current_values_)
    current_dtypes_ = set(map(lambda x: x.dtype.type, current_values_flat_))
    assert (
        len(current_dtypes_) == 1
    ), "multi-dtype not yet supported. TODO: manage multi-dtypes at Buffer level."
    current_dtype_ = current_dtypes_.pop()
    return current_dtype_



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Delay operator."""
import jax.numpy as jnp
from jax import tree_map

from wax.modules.buffer import Buffer


class Lag(Buffer):
    """Delay operator."""

    def __init__(self, lag, fill_value=jnp.nan, return_state=False, name=None):
        super().__init__(
            maxlen=lag + 1, fill_value=fill_value, return_state=return_state, name=name
        )
        self.lag = lag

    def __call__(self, x):
        buffer = super().__call__(x)
        return buffer[0]


# Helper functions


def tree_lag(shift=1):
    """Define a function which apply Lag module to a PyTree.
    This should be used inside a transformed function."""

    def apply_fn(*pytree):
        return tree_map(lambda x: Lag(shift)(x) if x is not None else None, pytree)

    return apply_fn



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unroll modules on data along first axis."""
import logging
import warnings
from collections import namedtuple
from functools import partial
from typing import Any, Callable, NamedTuple, Union, cast

import haiku as hk
import jax
import jax.numpy as jnp
from haiku import TransformedWithState
from jax._src.lax.control_flow import fori_loop
from jax.tree_util import tree_flatten, tree_map, tree_unflatten
from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


class UnrollTransformedWithState(NamedTuple):
    init: Callable
    apply: Callable


class ScanState(NamedTuple):
    fun_state: Any
    rng: jnp.ndarray


def unroll_transform_with_state(
    fun: Union[Callable, TransformedWithState, UnrollTransformedWithState],
    skip_first: bool = False,
    dynamic: bool = True,
    pbar: bool = False,
) -> UnrollTransformedWithState:
    """Transforms a function using Haiku modules into a pair of pure functions.
        which is unrolled on input arguments.

    Args:
        fun : callable or pair of pure functions (init, apply).
        skip_first : if true, first value of the sequence is not used in apply.
        dynamic : if true,  unroll using jax.lax.scan.
        pbar: if true, activate progress bar. Currently, it only works when dynamic=False.
    """
    if callable(fun):
        tfunc = hk.transform_with_state(fun)
    else:
        tfunc = cast(TransformedWithState, fun)
    del fun

    def init_fn(rng: jnp.ndarray, *args, **kwargs):
        xs = (args, kwargs)
        args_0, kwargs_0 = tree_map(lambda x: x[0], xs)
        params, state = tfunc.init(rng, *args_0, **kwargs_0)
        return params, state

    def apply_fn(params: Any, state: Any, rng: jnp.ndarray, *args, **kwargs):
        def scan_f(scan_state, inputs):
            state, rng = scan_state
            args_step, kwargs_step = inputs
            if rng is not None:
                (rng, sub_rng) = jax.random.split(rng)
            else:
                sub_rng = None
            outputs, state = tfunc.apply(
                params, state, sub_rng, *args_step, **kwargs_step
            )
            return ScanState(state, rng), outputs

        xs = (args, kwargs)

        if skip_first:
            xs = tree_map(lambda x: x[1:], xs)

        if dynamic:
            scan = jax.lax.scan
        else:
            scan = partial(static_scan, pbar=pbar)
        scan_state, output_sequence = scan(scan_f, init=ScanState(state, rng), xs=xs)
        final_state, final_rng = scan_state
        return output_sequence, final_state

    return UnrollTransformedWithState(init_fn, apply_fn)


def unroll(
    fun: Union[Callable, TransformedWithState, UnrollTransformedWithState],
    skip_first: bool = False,
    dynamic: bool = True,
    pbar: bool = False,
    return_final_state=False,
    rng=None,
    params=None,
    state=None,
):
    """Transforms a function using Haiku modules into a function
    which is unrolled on input arguments.

    Args:
        fun : callable or pair of pure functions (init, apply).
        skip_first : if true, first value of the sequence is not used in apply.
        dynamic : if true,  unroll using jax.lax.scan.
        pbar: if true, activate progress bar. Currently, it only works when dynamic=False.
        return_final_state : if true, the returned function return unrolled outputs and final state.
        rng : if specified, used as rng key for the init and apply functions.
        params : if specified, used as params for the apply function.
        state : if sepecified, used as initial state for the apply function.

    Returns:
        apply_fn: wrapped function.

    """
    if not isinstance(fun, UnrollTransformedWithState):
        fun = unroll_transform_with_state(
            fun, skip_first=skip_first, dynamic=dynamic, pbar=pbar
        )

    def apply_fn(*args, **kwargs):
        fun_init_params, fun_init_state = fun.init(rng, *args, **kwargs)
        init_params = fun_init_params if params is None else params
        init_state = fun_init_state if state is None else state
        output, final_state = fun.apply(init_params, init_state, rng, *args, **kwargs)
        if return_final_state:
            return output, final_state
        else:
            return output

    return apply_fn


def init_params_state(
    fun: hk.TransformedWithState,
    rng: jnp.ndarray,
    *args,
    **kwargs,
):
    """Call init of a TransformedUnrollWithState pair."""
    warnings.warn(
        "Deprecated function init_params_state. Use unroll_transform_with_state instead."
        "This function may be removed in the near future.",
        DeprecationWarning,
        2,
    )

    return unroll_transform_with_state(fun).init(rng, *args, **kwargs)


def dynamic_unroll(
    fun: Union[Callable, hk.TransformedWithState],
    params: Any,
    state: Any,
    rng: jnp.ndarray,
    skip_first: bool = False,
    *args,
    **kwargs,
):
    """Unroll a TransformedWithState function using jax.lax.scan.

    Args:
        fun : callable or pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if true, first value of the sequence is not used in apply.
        args, kwargs : Nested data structures with sequences as leaves passed to init and apply
            of the TransformedWithState pair.
    """
    warnings.warn(
        "Deprecated function dynamic_unroll. Use unroll or unroll_transform_with_state instead."
        "This function may be removed in the near future.",
        DeprecationWarning,
        2,
    )

    tfun = unroll_transform_with_state(fun, skip_first, dynamic=True)
    del fun

    fun_init_params, fun_init_state = tfun.init(rng, *args, **kwargs)
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state

    return tfun.apply(params, state, rng, *args, **kwargs)


def static_unroll(
    fun: hk.TransformedWithState,
    params: Any,
    state: Any,
    rng: jnp.ndarray,
    skip_first: bool = False,
    *args,
    pbar: bool = False,
    **kwargs,
):
    """Unroll a TransformedWithState function using static_scan.

    Args:
        fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if true, first value of the sequence is not used in apply.
        args, kwargs : Nested datastructure with sequences as leaves passed to init and apply
            of the TransformedWithState pair.
        pbar : if true, activate progress bar.
    """
    warnings.warn(
        "Deprecated function static_unroll. Use unroll or unroll_transform_with_state instead."
        "This function may be removed in the near future.",
        DeprecationWarning,
        2,
    )

    tfun = unroll_transform_with_state(fun, skip_first, dynamic=False, pbar=pbar)
    del fun

    fun_init_params, fun_init_state = tfun.init(rng, *args, **kwargs)
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state

    return tfun.apply(params, state, rng, *args, **kwargs)


def iter_first_axis(xs, pbar=False):
    """Iterate over first axis of a nested data structure."""
    xs_flat, treedef = tree_flatten(xs)
    T = len(xs_flat[0])
    if pbar:
        dates = tqdm(list(range(T)), desc="iter_first_axis (static_unroll)")
    else:
        dates = range(T)
    for t in dates:
        x_flat = map(lambda x: x[t], xs_flat)
        yield tree_unflatten(treedef, x_flat)


def static_scan(scan_f, init, xs=None, length=None, pbar=False):
    """Scan a function over leading array axes while carrying along state.

    Args:
        scan_f : function to apply
        init: initial value for first argument of scan_f
        xs : sequence of second argument of scan_f
        length: lenght of the output sequence.
        pbar: if true, activate a progress bar.

    Note:  Python implementation of jax.lax.scan
    See https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html#jax-lax-scan
    """
    if xs is None:
        xs = [None] * length
    carry = init
    ys = None
    for x in iter_first_axis(xs, pbar=pbar):
        carry, y = scan_f(carry, x)

        y, treedef = tree_flatten(y)
        if ys is None:
            ys = tuple([] for _ in y)
        for ys_, y_ in zip(ys, y):
            ys_.append(y_)

    ys = map(jnp.stack, ys)
    ys = tree_unflatten(treedef, ys)
    return carry, ys


# def static_scan_generator(scan_f, init, xs, length=None):
#     """Scan a function over leading array axes while carrying along state.
#
#     Python implementation of jax.lax.scan
#
#     See https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html#jax-lax-scan
#     """
#     if xs is None:
#         xs = [None] * length
#     carry = init
#     for x in iter_first_axis(xs):
#         carry, y = scan_f(carry, x)
#         yield carry, y
#
#
# def static_unroll_generator(
#     fun, params, initial_state, key, xs, skip_first=False, pbar=True
# ):
#     """Generator version of static_unroll"""
#     # inline static_unroll
#     def scan_f(prev_state, inputs):
#         outputs, next_state = fun.apply(fun_params, prev_state, key, inputs)
#         return next_state, outputs
#
#     # init
#     x0 = tree_map(lambda x: x[0], xs)
#     fun_params, fun_init_state = fun.init(key, x0)
#     init_state = fun_init_state if initial_state is None else initial_state
#     # apply
#     if skip_first:
#         xs = tree_map(lambda x: x[1:], xs)
#     for carry, y in static_scan_generator(scan_f, init=init_state, xs=xs):
#         yield carry, y


def dynamic_unroll_fori_loop(
    fun: hk.TransformedWithState,
    params: Any,
    state: Any,
    rng: jnp.ndarray,
    skip_first: bool = False,
    *args,
    **kwargs,
):
    """Unroll a TransformedWithState function using fori_loop.

    Args:
        fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if true, first value of the sequence is not used in apply.
        args, kwargs : Nested datastructure with sequences as leaves passed to init and apply
            of the TransformedWithState pair.
    """
    LoopState = namedtuple("LoopState", "params, state, rng, x, output_sequence")

    fun_init_params, fun_init_state = unroll_transform_with_state(fun).init(
        rng, *args, **kwargs
    )
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state
    xs = (args, kwargs)

    if skip_first:
        xs = tree_map(lambda x: x[1:], xs)

    # get template output
    args_0, kwargs_0 = tree_map(lambda x: x[0], xs)
    output_template, _ = fun.apply(params, state, rng, *args_0, **kwargs_0)

    T = len(tree_flatten(xs)[0][0])
    output_sequence = tree_map(
        lambda x: jnp.full((T,) + x.shape, jnp.nan, x.dtype), output_template
    )

    # run loop
    loop_state = LoopState(params, state, rng, xs, output_sequence)

    @jax.jit
    def body_fun(i, loop_state):
        params, prev_state, rng, inputs, output_sequence = loop_state
        args_i, kwargs_i = tree_map(lambda x: x[i], inputs)
        outputs, next_state = fun.apply(params, prev_state, rng, *args_i, **kwargs_i)
        output_sequence = output_sequence.at[i].set(outputs)
        return LoopState(
            params,
            next_state,
            rng,
            inputs,
            output_sequence,
        )

    loop_state = fori_loop(0, T, body_fun, loop_state)

    # format results
    final_state = loop_state.state
    output_sequence = loop_state.output_sequence
    return output_sequence, final_state


def data_unroll(data_gen, pbar=True):
    xs = None
    if pbar:
        data_gen = tqdm(data_gen, desc="data_unroll")
    for obs, info in data_gen:
        x = obs
        x, treedef = tree_flatten(x)
        if xs is None:
            # initialize xs
            xs = tuple([] for _ in x)
        for xs_, x_ in zip(xs, x):
            xs_.append(x_)
    xs = map(lambda x: jnp.stack(x, axis=0), xs)
    xs = tree_unflatten(treedef, xs)
    return xs


# def gym_dynamic_unroll(gym_fun, seq, data_generator, skip_first=True):
#     """Unroll a function over a data generator and random number generator."""
#     xs = data_unroll(data_generator)
#     output, gym_state = dynamic_unroll(gym_fun, xs, next(seq), skip_first=skip_first)
#     return output, gym_state


def gym_static_unroll(
    gym_fun, params, state, seq, skip_first, data_generator, pbar=False
):
    """Unroll a function over a data generator and random number generator.
    Args:
        gym_fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        seq: generator of random number keys.
        skip_first : if true, first value of the sequence is not used in apply.
        data_generator : generator for gym_fun arguments.
    """
    obs, info = next(data_generator)
    fun_init_params, fun_init_state = gym_fun.init(next(seq), obs)
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state

    ys = None
    if not skip_first:
        y, state = gym_fun.apply(params, state, next(seq), obs)
        y, treedef = tree_flatten(y)
        if ys is None:
            ys = tuple([] for _ in y)

    if pbar:
        data_generator = tqdm(data_generator, desc="gym_static_unroll")
    for obs, info in data_generator:
        y, state = gym_fun.apply(params, state, next(seq), obs)
        y, treedef = tree_flatten(y)
        if ys is None:
            ys = tuple([] for _ in y)
        for ys_, y_ in zip(ys, y):
            ys_.append(y_)

    ys = map(jnp.stack, ys)
    ys = tree_unflatten(treedef, ys)
    return ys, state
